from django.apps import apps
from django.contrib import admin

from comment.tests.base import BaseCommentTest
from comment.models import FlagInstance, ReactionInstance


class TestCommentAdmin(BaseCommentTest):
    def test_all_models_are_registered(self):
        app = apps.get_app_config('comment')
        models = app.get_models()

        for model in models:
            try:
                self.assertIs(
                    True,
                    admin.site.is_registered(model),
                    msg=f'Did you forget to register the "{model.__name__}" in the django-admin?')
            except AssertionError as exception:
                # these models have been registred as inlines in the admin.
                if model in [ReactionInstance, FlagInstance]:
                    continue
                else:
                    raise exception



import os

__version__ = '2.6.1'


def _get_version():
    parent_project_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    with open(os.path.join(parent_project_dir, 'VERSION')) as version_file:
        version = version_file.read().strip().strip('v')

    return version


def check_release():
    try:
        release = _get_version()
    except FileNotFoundError:
        return

    assert release == __version__, 'Current version does not match with manifest VERSION'


check_release()


default_app_config = 'comment.apps.CommentConfig'



from unittest.mock import patch

from comment.signals import adjust_flagged_comments
from comment.tests.base import BaseCommentSignalTest
from comment.models import Comment, Flag, FlagInstance, Reaction, ReactionInstance
from comment.conf import settings


class TestPostMigrate(BaseCommentSignalTest):
    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1)
    def test_adjust_flagged_comments(self):
        comment_1 = self.create_comment(self.content_object_1)
        comment_2 = self.create_comment(self.content_object_2)
        self.assertFalse(comment_1.is_flagged)
        self.assertFalse(comment_2.is_flagged)

        self.create_flag_instance(self.user_1, comment_1)
        self.create_flag_instance(self.user_2, comment_1)

        self.create_flag_instance(self.user_1, comment_2)
        self.create_flag_instance(self.user_2, comment_2)

        # flagged comment with wrong state will be adjusted
        comment_1.flag.state = comment_1.flag.UNFLAGGED
        comment_1.flag.save()
        comment_1.flag.refresh_from_db()
        self.assertEqual(comment_1.flag.count, 2)
        self.assertEqual(comment_1.flag.state, comment_1.flag.UNFLAGGED)
        self.assertFalse(comment_1.is_flagged)

        # flagged comment with right state => will be skipped
        comment_2.flag.refresh_from_db()
        self.assertEqual(comment_2.flag.count, 2)
        self.assertEqual(comment_2.flag.state, comment_2.flag.FLAGGED)
        self.assertTrue(comment_2.is_flagged)

        adjust_flagged_comments(self)
        comment_1.flag.refresh_from_db()
        self.assertEqual(comment_1.flag.count, 2)
        self.assertTrue(comment_1.is_flagged)


class TestPostSave(BaseCommentSignalTest):
    def test_reaction_signal(self):
        """Test reaction model instance is created when a comment is created"""
        parent_comment = self.create_comment(self.content_object_1)
        self.assertIsNotNone(Reaction.objects.get(comment=parent_comment))
        # 1 reaction instance is created for every comment
        self.assertEqual(Reaction.objects.count(), Comment.objects.count())

    def test_flag_signal(self):
        """Test flag model instance is created when a comment is created"""
        current_count = Flag.objects.count()
        parent_comment = self.create_comment(self.content_object_1)
        self.assertIsNotNone(Flag.objects.get(comment=parent_comment))
        # 1 flag instance is created for every comment
        self.assertEqual(Flag.objects.count(), current_count + 1)

    def test_increase_reaction_count(self):
        comment = self.comment
        self.assertEqual(comment.reaction.likes, 0)
        # reaction instance created
        reaction_instance = ReactionInstance.objects.create(
            reaction=comment.reaction, user=self.user_1, reaction_type=self.LIKE.value)
        comment.reaction.refresh_from_db()
        self.assertEqual(comment.reaction.likes, 1)
        self.assertEqual(comment.reaction.dislikes, 0)

        # edit reaction instance won't change reaction count
        reaction_instance.reaction_type = self.DISLIKE.value
        reaction_instance.save()
        comment.reaction.refresh_from_db()
        self.assertEqual(comment.reaction.likes, 1)
        self.assertEqual(comment.reaction.dislikes, 0)

    def test_increase_flag_count(self):
        self.assertEqual(self.comment.flag.count, 0)
        # instance created
        self.set_flag(self.user, self.comment, **self.flag_data)
        self.comment.flag.refresh_from_db()
        self.assertEqual(self.comment.flag.count, 1)
        # instance edited won't increase the flag count
        flag_instance = FlagInstance.objects.get(user=self.user, flag__comment=self.comment)
        self.assertIsNotNone(flag_instance)
        flag_instance.info = 'change value for test'
        flag_instance.save()
        self.comment.flag.refresh_from_db()
        self.assertEqual(self.comment.flag.count, 1)


class TestPostDelete(BaseCommentSignalTest):
    def test_reaction_decrease_count(self):
        """Test reaction count is decreased when an instance is deleted"""
        comment = self.comment
        instance = self.create_reaction_instance(self.user, self.comment, self.LIKE.name)
        comment.refresh_from_db()
        self.assertEqual(comment.likes, 1)
        instance.delete()
        comment.refresh_from_db()
        self.assertEqual(comment.likes, 0)

    def test_flag_decrease_count(self):
        """Test flag count is decreased when an instance is deleted"""
        data = self.flag_data
        comment = self.comment
        instance = self.create_flag_instance(self.user, comment, **data)
        comment.refresh_from_db()
        self.assertEqual(comment.flag.count, 1)
        instance.delete()
        comment.refresh_from_db()
        self.assertEqual(comment.flag.count, 0)



from abc import abstractmethod, ABCMeta
from functools import lru_cache

from django.contrib.contenttypes.models import ContentType
from django.core.validators import EmailValidator
from django.http import JsonResponse
from django.core.exceptions import ImproperlyConfigured, ValidationError

from comment.exceptions import CommentBadRequest
from comment.messages import ContentTypeError, ExceptionError, EmailError
from comment.utils import get_request_data
from comment.messages import ErrorMessage
from comment.conf import settings


class BaseValidatorMixin:
    __metaclass__ = ABCMeta
    api = False
    error = None

    def dispatch(self, request, *args, **kwargs):
        """
            let rest framework handle the exception to choose the right renderer
            validate method shall be called in the derived API class
        """
        if self.api:
            return super().dispatch(request, *args, **kwargs)
        try:
            self.validate(request)
        except CommentBadRequest as exc:
            return JsonResponse({'type': ExceptionError.ERROR_TYPE, 'detail': exc.detail}, status=400)
        return super().dispatch(request, *args, **kwargs)

    @abstractmethod
    def validate(self, request):
        pass


class ContentTypeValidator(BaseValidatorMixin):
    model_obj = None

    def validate_app_name(self, app_name):
        if not app_name:
            self.error = ContentTypeError.APP_NAME_MISSING
            raise CommentBadRequest(self.error)

        if not ContentType.objects.filter(app_label=app_name).exists():
            self.error = ContentTypeError.APP_NAME_INVALID.format(app_name=app_name)
            raise CommentBadRequest(self.error)
        return app_name

    def validate_model_name(self, model_name):
        if not model_name:
            self.error = ContentTypeError.MODEL_NAME_MISSING
            raise CommentBadRequest(self.error)
        return str(model_name).lower()

    def validate_content_type_object(self, app_name, model_name):
        try:
            ct_object = ContentType.objects.get(model=model_name, app_label=app_name)
        except ContentType.DoesNotExist:
            self.error = ContentTypeError.MODEL_NAME_INVALID.format(model_name=model_name)
            raise CommentBadRequest(self.error)
        return ct_object

    def validate_model_id(self, model_id):
        if not model_id:
            self.error = ContentTypeError.MODEL_ID_MISSING
            raise CommentBadRequest(self.error)
        try:
            model_id = int(model_id)
        except ValueError:
            self.error = ContentTypeError.ID_NOT_INTEGER.format(var_name='model', id=model_id)
            raise CommentBadRequest(self.error)
        return model_id

    def validate_model_object(self, app_name, model_name, model_id):
        ct_object = self.validate_content_type_object(app_name, model_name)
        model_class = ct_object.model_class()
        model_query = model_class.objects.filter(id=model_id)
        if not model_query.exists() and model_query.count() != 1:
            self.error = ContentTypeError.MODEL_ID_INVALID.format(model_id=model_id, model_name=model_name)
            raise CommentBadRequest(self.error)
        return model_query.first()

    def validate(self, request):
        super().validate(request)
        app_name = get_request_data(request, 'app_name', self.api)
        model_name = get_request_data(request, 'model_name', self.api)
        model_id = get_request_data(request, 'model_id', self.api)
        validated_app_name = self.validate_app_name(app_name)
        validated_model_name = self.validate_model_name(model_name)
        validated_model_id = self.validate_model_id(model_id)
        self.model_obj = self.validate_model_object(validated_app_name, validated_model_name, validated_model_id)


class ParentIdValidator(BaseValidatorMixin):
    parent_comment = None

    def validate_parent_id(self, parent_id):
        try:
            parent_id = int(parent_id)
        except ValueError:
            self.error = ContentTypeError.ID_NOT_INTEGER.format(var_name='parent', id=parent_id)
            raise CommentBadRequest(self.error)
        return parent_id

    def validate_comment_object(self, model_id, parent_id):
        from comment.models import Comment

        try:
            comment = Comment.objects.get(id=parent_id, object_id=model_id)
        except Comment.DoesNotExist:
            self.error = ContentTypeError.PARENT_ID_INVALID.format(parent_id=parent_id)
            raise CommentBadRequest(self.error)
        return comment

    def validate(self, request):
        super().validate(request)
        model_id = get_request_data(request, 'model_id', self.api)
        parent_id = get_request_data(request, 'parent_id', self.api)
        if not parent_id or parent_id == '0':
            return
        validated_parent_id = self.validate_parent_id(parent_id)
        self.parent_comment = self.validate_comment_object(model_id, validated_parent_id)


class ValidatorMixin(ContentTypeValidator, ParentIdValidator):
    pass


class DABEmailValidator(EmailValidator):
    def __init__(self, email):
        super().__init__(EmailError.EMAIL_INVALID)
        self.email = email

    def is_valid(self):
        try:
            self(self.email)
            return True
        except ValidationError:
            return False


@lru_cache()
def _get_allowed_orders():
    allowed_orders = ['reaction__likes', 'reaction__dislikes', 'posted']
    # adds support for descending order as well
    allowed_orders.extend(list(map(lambda a: '-' + a, allowed_orders)))
    # django allows this to support random order
    allowed_orders.append('?')
    return allowed_orders


def _validate_orders_are_unique(preferred_orders):
    unique_values = set(map(lambda a: a.replace('-', ''), preferred_orders))
    if len(unique_values) != len(preferred_orders):
        duplicated_orders = list(set(preferred_orders) - unique_values)
        raise ImproperlyConfigured(
            ErrorMessage.DUPLICATE_ORDER_VALUE.format(
                duplicates=duplicated_orders, order=duplicated_orders[0].replace('-', '')
            )
        )


def _validate_order():
    preferred_orders = settings.COMMENT_ORDER_BY
    _validate_orders_are_unique(preferred_orders)

    allowed_orders = _get_allowed_orders()
    for preferred_order in preferred_orders:
        if preferred_order not in allowed_orders:
            raise ImproperlyConfigured(
                ErrorMessage.INVALID_ORDER_ARGUMENT.format(order=preferred_order, allowed_orders=allowed_orders)
            )
    return preferred_orders



from django.shortcuts import render
from django.template.loader import render_to_string
from django.test import RequestFactory

from comment.tests.base import BaseCommentTest


class InternationalizationTest(BaseCommentTest):
    """Testing values are not necessarily returned from a view functions"""

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        comment = cls.create_comment(cls.content_object_1)
        cls.context = {'comment': comment}

    def setUp(self):
        super().setUp()
        factory = RequestFactory()
        self.request = factory.get('/')

    def test_confirmation_email_template_text(self):
        response = render_to_string(template_name='comment/anonymous/confirmation_request.txt', context=self.context)
        self.assertHtmlTranslated(response)

    def test_confirmation_email_template_html(self):
        response = render(
            self.request,
            template_name='comment/anonymous/confirmation_request.html',
            context=self.context
        )
        self.assertHtmlTranslated(response.content)



# Generated by Django 3.1.4 on 2020-12-21 12:26

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('comment', '0010_auto_20201023_1442'),
    ]

    operations = [
        migrations.CreateModel(
            name='Follower',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email', models.EmailField(max_length=254)),
                ('username', models.CharField(max_length=50)),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
            ],
        ),
    ]



from unittest import skipIf
from unittest.mock import patch

from django.test import TestCase

from comment import __version__, _get_version, check_release


SKIP = False
try:
    MANIFEST_VERSION = _get_version()
except FileNotFoundError:
    SKIP = True


class TestVersion(TestCase):

    @skipIf(SKIP, 'Test skipped: development purpose only')
    def test_get_version(self):
        self.assertEqual(_get_version(), MANIFEST_VERSION)

    @skipIf(SKIP, 'Test skipped: development purpose only')
    def test_manifest_matches_init_version(self):
        self.assertEqual(__version__, MANIFEST_VERSION)

    @patch('comment._get_version', side_effect=[FileNotFoundError])
    def test_check_release_file_not_found(self, mocked_get_version):
        self.assertIsNone(check_release())
        mocked_get_version.assert_called_once()

    @skipIf(SKIP, 'Test skipped: development purpose only')
    @patch('comment.__version__', return_value='not_match')
    def test_check_release_raise_assertion_error_when_versions_does_not_match(self, _):
        self.assertRaises(AssertionError, check_release)

    def test_check_release_exit_with_no_error(self):
        self.assertIsNone(check_release())



from unittest import TestCase
from unittest.mock import patch

from django.core.exceptions import ImproperlyConfigured
from django.utils import timezone
from django.core import signing
from django.contrib.auth.models import AnonymousUser

from comment.conf import settings
from comment.utils import (
    get_model_obj, has_valid_profile, id_generator, get_comment_from_key, get_user_for_request, CommentFailReason,
    get_gravatar_img, get_profile_instance, is_comment_moderator, is_comment_admin, get_wrapped_words_number
)
from comment.tests.base import BaseCommentUtilsTest, Comment, RequestFactory
from comment.messages import ErrorMessage


class CommentUtilsTest(BaseCommentUtilsTest):
    def test_get_model_object(self):
        data = {
            'app_name': 'post',
            'model_name': 'Post',
            'model_id': self.post_1.id
        }
        model_object = get_model_obj(**data)
        self.assertIsNotNone(model_object)
        self.assertIsInstance(model_object, self.post_1.__class__)

    def test_get_gravatar_img(self):
        with patch.object(settings, 'COMMENT_USE_GRAVATAR', True):
            # email is not provided
            default_profile_pic_loc = settings.COMMENT_DEFAULT_PROFILE_PIC_LOC
            self.assertEqual(get_gravatar_img(''), default_profile_pic_loc)

            # email is provided
            self.assertTrue(get_gravatar_img('test').startswith('https://www.gravatar.com/avatar/'))

        # gravatar is disabled
        with patch.object(settings, 'COMMENT_USE_GRAVATAR', False):
            self.assertEqual(get_gravatar_img(''), default_profile_pic_loc)

    def test_get_profile_instance(self):
        # wrong content type
        with patch.object(settings, 'PROFILE_MODEL_NAME', 'wrong'):
            self.assertIsNone(get_profile_instance(self.user_1))

        # correct data
        with patch.object(settings, 'PROFILE_MODEL_NAME', 'userprofile'):
            self.assertIsNotNone(get_profile_instance(self.user_1))

        # profile model has no user related model
        with patch.object(settings, 'PROFILE_MODEL_NAME', None):
            self.assertIsNone(get_profile_instance(self.user_1))

    @patch.object(settings, 'COMMENT_USE_GRAVATAR', False)
    def test_has_valid_profile(self):
        with patch.object(settings, 'PROFILE_APP_NAME', 'user_profile'):
            with patch.object(settings, 'PROFILE_MODEL_NAME', 'userprofile'):
                self.assertIs(has_valid_profile(), True)

            # settings attr provided, profile model has no image
            with patch('comment.utils.hasattr', return_value=False):
                self.assertIs(has_valid_profile(), False)

            # one of settings attribute is missing
            with patch.object(settings, 'PROFILE_MODEL_NAME', ''):
                self.assertIs(has_valid_profile(), False)

            # settings attr provided with wrong value
            with patch.object(settings, 'PROFILE_MODEL_NAME', 'wrong_value'):
                self.assertIs(has_valid_profile(), False)

            with patch.object(settings, 'COMMENT_USE_GRAVATAR', True):
                self.assertIs(has_valid_profile(), True)

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', False)
    def test_is_comment_moderator_no_moderation(self):
        self.assertFalse(is_comment_moderator(self.moderator))

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', False)
    def test_is_comment_admin_no_moderation(self):
        self.assertFalse(is_comment_admin(self.admin))

    def test_user_for_request(self):
        request = self.factory.get('/')
        request.user = AnonymousUser()
        # test unauthenticated user
        self.assertIsNone(get_user_for_request(request))
        # test authenticated user
        request.user = self.user_1
        self.assertEqual(get_user_for_request(request), self.user_1)

    @patch.object(settings, 'COMMENT_WRAP_CONTENT_WORDS', None)
    def test_get_wrapped_words_number_return_0_for_None(self):
        self.assertEqual(get_wrapped_words_number(), 0)

    @patch.object(settings, 'COMMENT_WRAP_CONTENT_WORDS', 'test')
    def test_get_wrapped_words_number_fails_on_non_int_value(self):
        with self.assertRaises(ImproperlyConfigured) as e:
            get_wrapped_words_number()
        self.assertEqual(str(e.exception), ErrorMessage.WRAP_CONTENT_WORDS_NOT_INT)
        self.assertTextTranslated(ErrorMessage.WRAP_CONTENT_WORDS_NOT_INT)

    @patch.object(settings, 'COMMENT_WRAP_CONTENT_WORDS', 20)
    def test_get_wrapped_words_number_return_specified_setting_value(self):
        self.assertEqual(get_wrapped_words_number(), 20)


class BaseAnonymousCommentTest(BaseCommentUtilsTest):
    def setUp(self):
        super().setUp()
        self.time_posted = timezone.now()
        _email = 'test-1@acme.edu'
        _content = 'posting anonymous comment'
        _parent = None
        _factory = RequestFactory()
        self.comment_obj = Comment(
            content_object=self.post_1,
            content=_content,
            user=None,
            parent=_parent,
            email=_email,
            posted=self.time_posted
        )

        self.key = signing.dumps(self.comment_obj.to_dict(), compress=True)
        self.request = _factory.get('/')


class TestGetCommentFromKey(BaseAnonymousCommentTest):
    def test_bad_signature(self):
        key = self.key + 'invalid'
        response = get_comment_from_key(key)

        self.assertEqual(response.is_valid, False)
        self.assertEqual(response.why_invalid, CommentFailReason.BAD)
        self.assertIsNone(response.obj)

    def test_key_error(self):
        comment_dict = self.comment_obj.to_dict().copy()
        comment_dict.pop('model_name')
        key = signing.dumps(comment_dict)
        response = get_comment_from_key(key)

        self.assertEqual(response.is_valid, False)
        self.assertEqual(response.why_invalid, CommentFailReason.BAD)
        self.assertIsNone(response.obj)

    def test_attribute_error(self):
        comment_dict = self.comment_obj.to_dict().copy()
        comment_dict['model_name'] = 1
        key = signing.dumps(comment_dict)
        response = get_comment_from_key(key)

        self.assertEqual(response.is_valid, False)
        self.assertEqual(response.why_invalid, CommentFailReason.BAD)
        self.assertIsNone(response.obj)

    def test_value_error(self):
        comment_dict = self.comment_obj.to_dict().copy()
        comment_dict['user'] = 1
        key = signing.dumps(comment_dict)
        response = get_comment_from_key(key)

        self.assertEqual(response.is_valid, False)
        self.assertEqual(response.why_invalid, CommentFailReason.BAD)
        self.assertIsNone(response.obj)

    def test_comment_exists(self):
        comment_dict = self.comment_obj.to_dict().copy()
        comment = self.create_anonymous_comment(posted=timezone.now(), email='a@a.com')
        comment_dict.update({
            'posted': str(comment.posted),
            'email': comment.email
        })
        key = signing.dumps(comment_dict)
        response = get_comment_from_key(key)

        self.assertEqual(response.is_valid, False)
        self.assertEqual(response.why_invalid, CommentFailReason.EXISTS)
        self.assertIsNone(response.obj)

    def test_success(self):
        response = get_comment_from_key(self.key)

        self.assertEqual(response.is_valid, True)
        self.assertEqual(response.why_invalid, None)
        self.assertIsInstance(response.obj, Comment)


class UtilsTest(TestCase):
    """Test general purpose utilities that aren't necessarily related to a comment"""

    def setUp(self):
        self.len_id = 6

    def test_id_generator_length(self):
        self.assertEqual(self.len_id, len(id_generator()))

    def test_id_generator_generates_different_ids(self):
        self.assertNotEqual(id_generator(), id_generator())

    def test_id_generator_prefix(self):
        prefix = 'comment'
        output = id_generator(prefix=prefix)
        self.assertEqual(True, output.startswith(prefix))
        self.assertEqual(self.len_id + len(prefix), len(output))

    def test_id_generator_suffix(self):
        suffix = 'comment'
        output = id_generator(suffix=suffix)
        self.assertEqual(True, output.endswith(suffix))
        self.assertEqual(self.len_id + len(suffix), len(output))

    def test_id_generator_chars(self):
        import string   # flake8:no qa
        chars = string.ascii_uppercase
        output = id_generator(chars=chars)
        self.assertEqual(output, output.upper())

    def test_id_generator_len(self):
        len_id = 8
        self.assertEqual(len_id, len(id_generator(len_id=len_id)))



from math import ceil

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from django.db import models
from django.utils import timezone

from comment.managers import CommentManager
from comment.conf import settings
from comment.utils import is_comment_moderator


class Comment(models.Model):
    user = models.ForeignKey(get_user_model(), on_delete=models.CASCADE, blank=True, null=True)
    email = models.EmailField(blank=True)
    parent = models.ForeignKey('self', on_delete=models.CASCADE, null=True, blank=True)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')
    content = models.TextField()
    urlhash = models.CharField(
        max_length=50,
        unique=True,
        editable=False
        )
    posted = models.DateTimeField(default=timezone.now, editable=False)
    edited = models.DateTimeField(auto_now=True)

    objects = CommentManager()

    class Meta:
        ordering = ['-posted', ]

    def __str__(self):
        username = self.get_username()
        _content = self.content[:20]
        if not self.parent:
            return f'comment by {username}: {_content}'
        else:
            return f'reply by {username}: {_content}'

    def __repr__(self):
        return self.__str__()

    def to_dict(self):
        return {
            'user': self.user,
            'content': self.content,
            'email': self.email,
            'posted': str(self.posted),
            'app_name': self.content_type.app_label,
            'model_name': self.content_type.model,
            'model_id': self.object_id,
            'parent': getattr(self.parent, 'id', None)
        }

    def _get_reaction_count(self, reaction_type):
        return getattr(self.reaction, reaction_type, None)

    def replies(self, include_flagged=False):
        manager = self.__class__.objects
        if include_flagged:
            qs = manager.all()
        else:
            qs = manager.all_exclude_flagged()

        return manager._filter_parents(qs, parent=self)

    def _set_unique_urlhash(self):
        if not self.urlhash:
            self.urlhash = self.__class__.objects.generate_urlhash()
            while self.__class__.objects.filter(urlhash=self.urlhash).exists():
                self.urlhash = self.__class__.objects.generate_urlhash()

    def _set_email(self):
        if self.user:
            self.email = getattr(self.user, self.user.EMAIL_FIELD, '')

    def _get_username_for_anonymous(self):
        if settings.COMMENT_USE_EMAIL_FIRST_PART_AS_USERNAME:
            return self.email.split('@')[0]

        return settings.COMMENT_ANONYMOUS_USERNAME

    def get_username(self):
        user = self.user
        if not user:
            return self._get_username_for_anonymous()

        return getattr(user, user.USERNAME_FIELD)

    def save(self, *args, **kwargs):
        self._set_unique_urlhash()
        self._set_email()
        super(Comment, self).save(*args, **kwargs)

    def get_url(self, request):
        page_url = self.content_object.get_absolute_url()
        comments_per_page = settings.COMMENT_PER_PAGE
        if comments_per_page:
            qs_all_parents = self.__class__.objects.filter_parents_by_object(
                self.content_object, include_flagged=is_comment_moderator(request.user)
                )
            position = qs_all_parents.filter(posted__gte=self.posted).count() + 1
            if position > comments_per_page:
                page_url += '?page=' + str(ceil(position / comments_per_page))
        return page_url + '#' + self.urlhash

    @property
    def is_parent(self):
        return self.parent is None

    @property
    def is_edited(self):
        if self.user:
            return self.posted.timestamp() + 1 < self.edited.timestamp()
        return False

    @property
    def likes(self):
        return self._get_reaction_count('likes')

    @property
    def dislikes(self):
        return self._get_reaction_count('dislikes')

    @property
    def is_flagged(self):
        if hasattr(self, 'flag') and self.flag.is_flag_enabled:
            return self.flag.state != self.flag.UNFLAGGED
        return False

    @property
    def has_flagged_state(self):
        if hasattr(self, 'flag'):
            return self.flag.state == self.flag.FLAGGED
        return False

    @property
    def has_rejected_state(self):
        if hasattr(self, 'flag'):
            return self.flag.state == self.flag.REJECTED
        return False

    @property
    def has_resolved_state(self):
        if hasattr(self, 'flag'):
            return self.flag.state == self.flag.RESOLVED
        return False



from comment.tests.base import BaseCommentMigrationTest


class CommentMigrationTest(BaseCommentMigrationTest):
    migrate_from = '0007_auto_20200620_1259'
    migrate_to = '0009_auto_20200811_1945'

    def create_comment(self):
        self.instance += 1
        return self.old_model.objects.create(
            content_type_id=self.ct_object.id,
            object_id=self.instance,
            content=f'test migration - {self.instance}',
            user_id=self.user.id,
        )

    def setUpBeforeMigration(self, apps):
        self.instance = 0
        self.comment = self.create_comment()

    def test_email_and_urlhash_migrated(self):
        comment = self.new_model.objects.get(id=self.comment.id)

        self.assertEqual(hasattr(comment, 'urlhash'), True)
        self.assertEqual(comment.email, comment.user.email)



from django.views.generic.base import View

from comment.models import Follower
from comment.mixins import CanSubscribeMixin, DABResponseData
from comment.responses import UTF8JsonResponse
from comment.validators import ContentTypeValidator, DABEmailValidator
from comment.messages import FollowError


class BaseToggleFollowView(ContentTypeValidator, DABResponseData):
    response_class = None

    def get_response_class(self):
        assert self.response_class is not None, (
                "'%s' should either include a `response_class` attribute, "
                "or override the `get_response_class()` method."
                % self.__class__.__name__
        )
        return self.response_class

    def post(self, request, *args, **kwargs):
        """ Allow authenticated users only, anonymous may be added in the future """
        email = request.POST.get('email', None)
        response_class = self.get_response_class()
        if email and not DABEmailValidator(email).is_valid():
            self.error = {
                'invalid_email': DABEmailValidator.message
            }
            self.status = 400
            return response_class(self.json(), status=self.status)

        user = request.user
        if user.email:
            email = user.email

        if not email:
            self.error = {
                'email_required': FollowError.EMAIL_REQUIRED.format(model_object=str(self.model_obj))
            }
            self.status = 400
            return response_class(self.json(), status=self.status)

        if not user.email:
            user.email = email
            user.save()

        username = user.username or email.aplit('@')[0]
        following = Follower.objects.toggle_follow(email=email, model_object=self.model_obj, username=username)
        self.data = {
            'following': following,
            'app_name': self.model_obj._meta.app_label,
            'model_name': self.model_obj.__class__.__name__,
            'model_id': self.model_obj.id,
            'model_object': str(self.model_obj)
        }
        return response_class(self.json())


class ToggleFollowView(BaseToggleFollowView, CanSubscribeMixin, View):
    response_class = UTF8JsonResponse



from django.urls import reverse
from rest_framework import status

from comment.tests.base import BaseCommentViewTest
from comment.messages import ReactionInfo
from comment.conf import settings


class SetReactionViewTest(BaseCommentViewTest):
    def setUp(self):
        super().setUp()
        self.comment = self.create_comment(self.content_object_1)

    @staticmethod
    def get_reaction_url(obj_id, action):
        return reverse('comment:react', kwargs={
            'pk': obj_id,
            'reaction': action
        })

    def test_set_reaction_for_authenticated_users(self):
        """Test whether users can create/change reactions using view"""
        _url = self.get_reaction_url(self.comment.id, 'like')
        response = self.client.post(_url)
        data = {
            'data': {
                'status': 0,
                'likes': 1,
                'dislikes': 0,
            },
            'msg': ReactionInfo.UPDATED_SUCCESS,
            'anonymous': False,
            'error': None,
        }
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        server_response = response.json()
        self.assertDictEqual(server_response, data)
        self.assertTextTranslated(server_response['msg'], _url)

    def test_set_reaction_for_old_comments(self):
        """Test backward compatibility for this update"""
        _url = self.get_reaction_url(self.comment.id, 'like')
        # delete the reaction object
        self.comment.reaction.delete()
        response = self.client.post(_url)
        data = {
            'data': {
                'status': 0,
                'likes': 1,
                'dislikes': 0,
            },
            'msg': ReactionInfo.UPDATED_SUCCESS,
            'anonymous': False,
            'error': None,
        }
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        server_response = response.json()
        self.assertDictEqual(server_response, data)
        self.assertTextTranslated(server_response['msg'], _url)

    def test_set_reaction_for_unauthenticated_users(self):
        """Test whether unauthenticated users can create/change reactions using view"""
        _url = self.get_reaction_url(self.comment.id, 'dislike')
        self.client.logout()
        response = self.client.post(_url)
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.url, '{}?next={}'.format(settings.LOGIN_URL, _url))

    def test_get_request(self):
        """Test whether GET requests are allowed or not"""
        _url = self.get_reaction_url(self.comment.id, 'like')
        response = self.client.get(_url)
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_non_ajax_requests(self):
        """Test response if non AJAX requests are sent"""
        _url = self.get_reaction_url(self.comment.id, 'like')
        response = self.client_non_ajax.post(_url)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_incorrect_comment_id(self):
        """Test response when an incorrect comment id is passed"""
        _url = self.get_reaction_url(102_876, 'like')
        response = self.client.post(_url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_incorrect_reaction(self):
        """Test response when incorrect reaction is passed"""
        _url = self.get_reaction_url(self.comment.id, 'likes')
        response = self.client.post(_url)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # test incorrect type
        _url = self.get_reaction_url(self.comment.id, 1)
        response = self.client.post(_url)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)



from django.core.exceptions import ValidationError
from rest_framework import generics, permissions, status
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response
from rest_framework.views import APIView

from comment.validators import ValidatorMixin, ContentTypeValidator
from comment.api.serializers import CommentSerializer, CommentCreateSerializer
from comment.api.permissions import (
    IsOwnerOrReadOnly, FlagEnabledPermission, CanChangeFlaggedCommentState,
    SubscriptionEnabled, CanGetSubscribers)
from comment.models import Comment, Reaction, ReactionInstance, Flag, FlagInstance, Follower
from comment.utils import get_comment_from_key, CommentFailReason
from comment.messages import FlagError, EmailError
from comment.views import BaseToggleFollowView
from comment.mixins import CommentCreateMixin


class CommentCreate(ValidatorMixin, generics.CreateAPIView):
    serializer_class = CommentCreateSerializer
    api = True

    def get_serializer_context(self):
        self.validate(self.request)
        context = super().get_serializer_context()
        context['model_obj'] = self.model_obj
        context['parent_comment'] = self.parent_comment
        context['email'] = self.request.GET.get('email', None)
        return context


class CommentList(ContentTypeValidator, generics.ListAPIView):
    serializer_class = CommentSerializer
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)
    api = True

    def get_queryset(self):
        self.validate(self.request)
        return Comment.objects.filter_parents_by_object(self.model_obj)


class CommentDetail(generics.RetrieveUpdateDestroyAPIView):
    queryset = Comment.objects.all()
    serializer_class = CommentSerializer
    permission_classes = (permissions.IsAuthenticatedOrReadOnly, IsOwnerOrReadOnly)


class CommentDetailForReaction(generics.RetrieveAPIView):
    queryset = Comment.objects.all()
    serializer_class = CommentSerializer
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['reaction_update'] = True
        return context

    def post(self, request, *args, **kwargs):
        comment = get_object_or_404(Comment, id=kwargs.get('pk'))
        reaction_type = kwargs.get('reaction', None)
        reaction_obj = Reaction.objects.get_reaction_object(comment)
        try:
            ReactionInstance.objects.set_reaction(
                user=request.user,
                reaction=reaction_obj,
                reaction_type=reaction_type
            )
        except ValidationError as e:
            return Response({'detail': e.messages}, status=status.HTTP_400_BAD_REQUEST)

        comment.reaction.refresh_from_db()
        serializer = self.get_serializer(comment)
        return Response(serializer.data, status=status.HTTP_200_OK)


class CommentDetailForFlag(generics.RetrieveAPIView):
    queryset = Comment.objects.all()
    serializer_class = CommentSerializer
    permission_classes = (permissions.IsAuthenticatedOrReadOnly, FlagEnabledPermission)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['flag_update'] = True
        return context

    def post(self, request, *args, **kwargs):
        comment = get_object_or_404(Comment, id=kwargs.get('pk'))
        flag = Flag.objects.get_for_comment(comment)
        reason = request.data.get('reason') or request.POST.get('reason')
        info = request.data.get('info') or request.POST.get('info')
        try:
            FlagInstance.objects.set_flag(request.user, flag, reason=reason, info=info)
        except ValidationError as e:
            return Response({'detail': e.messages}, status=status.HTTP_400_BAD_REQUEST)

        serializer = self.get_serializer(comment)
        return Response(serializer.data, status=status.HTTP_200_OK)


class CommentDetailForFlagStateChange(generics.RetrieveAPIView):
    queryset = Comment.objects.all()
    serializer_class = CommentSerializer
    permission_classes = (CanChangeFlaggedCommentState, )

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['flag_update'] = True
        return context

    def post(self, request, *args, **kwargs):
        comment = get_object_or_404(Comment, id=kwargs.get('pk'))
        flag = Flag.objects.get_for_comment(comment)
        if not comment.is_flagged:
            return Response(
                {'detail': FlagError.REJECT_UNFLAGGED_COMMENT},
                status=status.HTTP_400_BAD_REQUEST
            )
        state = request.data.get('state') or request.POST.get('state')
        try:
            state = flag.get_clean_state(state)
            if not comment.is_edited and state == flag.RESOLVED:
                return Response(
                    {'detail': FlagError.RESOLVE_UNEDITED_COMMENT},
                    status=status.HTTP_400_BAD_REQUEST
                )
            flag.toggle_state(state, request.user)
        except ValidationError as e:
            return Response({'detail': e.messages}, status=status.HTTP_400_BAD_REQUEST)

        serializer = self.get_serializer(comment)
        return Response(serializer.data, status=status.HTTP_200_OK)


class ConfirmComment(APIView, CommentCreateMixin):
    def get(self, request, *args, **kwargs):
        key = kwargs.get('key', None)
        temp_comment = get_comment_from_key(key)

        if temp_comment.why_invalid == CommentFailReason.BAD:
            return Response({'detail': EmailError.BROKEN_VERIFICATION_LINK}, status=status.HTTP_400_BAD_REQUEST)

        if temp_comment.why_invalid == CommentFailReason.EXISTS:
            return Response({'detail': EmailError.USED_VERIFICATION_LINK}, status=status.HTTP_200_OK)

        comment = self.perform_save(temp_comment.obj, request)

        return Response(CommentSerializer(comment).data, status=status.HTTP_201_CREATED)


class ToggleFollowAPI(BaseToggleFollowView, APIView):
    api = True
    response_class = Response
    permission_classes = (SubscriptionEnabled, permissions.IsAuthenticated)

    def post(self, request, *args, **kwargs):
        self.validate(request)
        return super().post(request, *args, **kwargs)


class SubscribersAPI(ContentTypeValidator, APIView):
    api = True
    permission_classes = (CanGetSubscribers,)

    def get(self, request, *args, **kwargs):
        self.validate(request)
        return Response({
            'app_name': self.model_obj._meta.app_label,
            'model_name': self.model_obj.__class__.__name__,
            'model_id': self.model_obj.id,
            'followers': Follower.objects.get_emails_for_model_object(self.model_obj)
        })



from unittest.mock import patch

from django.http import JsonResponse
from django.test import TestCase
from django.views import View
from django.core.exceptions import ImproperlyConfigured
from rest_framework import status
from rest_framework.generics import ListAPIView

from comment.tests.base import BaseCommentMixinTest
from comment.validators import CommentBadRequest, ValidatorMixin, _validate_order
from comment.conf import settings
from comment.messages import ExceptionError, ContentTypeError


class MockedContentTypeValidatorView(ValidatorMixin, View):
    api = False


class MockedContentTypeValidatorAPIView(ValidatorMixin, ListAPIView):
    api = True

    def get(self, request, *args, **kwargs):
        self.validate(request)
        return JsonResponse({})


class CustomValidationTest(TestCase):
    def test_can_create_custom_validation(self):
        # no params
        validator = CommentBadRequest()
        self.assertEqual(validator.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(validator.detail, ExceptionError.BAD_REQUEST)

        # with params
        validator = CommentBadRequest(detail='Not Found', status_code=404)
        self.assertEqual(validator.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(validator.detail, "Not Found")


class ValidatorMixinTest(BaseCommentMixinTest):

    def setUp(self):
        super().setUp()
        self.view = MockedContentTypeValidatorView()
        self.base_url = '/'

    def test_missing_app_name(self):
        url_data = self.data.copy()
        url_data.pop('app_name')
        request = self.factory.get(self.get_url(**url_data))
        response = self.view.dispatch(request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(self.view.error, ContentTypeError.APP_NAME_MISSING)

    def test_missing_model_type(self):
        url_data = self.data.copy()
        url_data.pop('model_name')
        request = self.factory.get(self.get_url(**url_data))
        response = self.view.dispatch(request)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(self.view.error, ContentTypeError.MODEL_NAME_MISSING)

    def test_missing_model_id(self):
        url_data = self.data.copy()
        url_data.pop('model_id')
        request = self.factory.get(self.get_url(**url_data))
        response = self.view.dispatch(request)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(self.view.error, ContentTypeError.MODEL_ID_MISSING)

    def test_invalid_model_name(self):
        url_data = self.data.copy()
        model_name = 'not exists'
        url_data['model_name'] = model_name
        request = self.factory.get(self.get_url(**url_data))
        response = self.view.dispatch(request)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(self.view.error, ContentTypeError.MODEL_NAME_INVALID.format(model_name=model_name))

    def test_invalid_app_name(self):
        url_data = self.data.copy()
        app_name = 'not exists'
        url_data['app_name'] = app_name
        request = self.factory.get(self.get_url(**url_data))
        response = self.view.dispatch(request)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(self.view.error, ContentTypeError.APP_NAME_INVALID.format(app_name=app_name))

    def test_model_id_does_not_exist(self):
        url_data = self.data.copy()
        model_id = 100
        url_data['model_id'] = model_id
        request = self.factory.get(self.get_url(**url_data))
        response = self.view.dispatch(request)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            self.view.error,
            ContentTypeError.MODEL_ID_INVALID.format(model_id=model_id, model_name=url_data['model_name'])
        )

    def test_model_id_non_integral(self):
        url_data = self.data.copy()
        model_id = 'not integral'
        url_data['model_id'] = model_id
        request = self.factory.get(self.get_url(**url_data))
        response = self.view.dispatch(request)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(self.view.error, ContentTypeError.ID_NOT_INTEGER.format(var_name='model', id=model_id))

    def test_not_valid_parent_id(self):
        url_data = self.data.copy()
        parent_id = 1000
        url_data['parent_id'] = parent_id
        request = self.factory.get(self.get_url(**url_data))
        response = self.view.dispatch(request)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(self.view.error, ContentTypeError.PARENT_ID_INVALID.format(parent_id=parent_id))

    def test_parent_id_no_int(self):
        url_data = self.data.copy()
        parent_id = 'string'
        url_data['parent_id'] = parent_id
        request = self.factory.get(self.get_url(**url_data))
        response = self.view.dispatch(request)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(self.view.error, ContentTypeError.ID_NOT_INTEGER.format(var_name='parent', id=parent_id))

    def test_api_case_success(self):
        view = MockedContentTypeValidatorAPIView()
        request = self.factory.get(self.get_url(**self.data))
        response = view.dispatch(request)
        self.assertEqual(response.status_code, 200)

    def test_api_case_missing_app_name(self):
        view = MockedContentTypeValidatorAPIView()
        url_data = self.data.copy()
        url_data.pop('app_name')
        request = self.factory.get(self.get_url(**url_data))
        response = view.dispatch(request)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['detail'], view.error)


class ValidateOrderTest(TestCase):
    def test_success(self):
        order = ['-reaction__likes']
        with patch.object(settings, 'COMMENT_ORDER_BY', order):
            self.assertEqual(_validate_order(), order)

    def test_incorrect_value_raises_exception(self):
        order = ['err']
        with patch.object(settings, 'COMMENT_ORDER_BY', order):
            with self.assertRaises(ImproperlyConfigured) as error:
                _validate_order()

        exception = error.exception

        self.assertIsInstance(exception, ImproperlyConfigured)
        self.assertIs(order[0] in str(exception), True)

    def test_duplicate_value_raises_exception(self):
        order = ['posted', '-posted']
        with patch.object(settings, 'COMMENT_ORDER_BY', order):
            with self.assertRaises(ImproperlyConfigured) as error:
                _validate_order()

        exception = error.exception

        self.assertIsInstance(exception, ImproperlyConfigured)
        self.assertIs(order[0] in str(exception), True)



from django.shortcuts import render, get_object_or_404, redirect
from django.template.loader import render_to_string
from django.utils import timezone
from django.contrib import messages

from comment.models import Comment
from comment.forms import CommentForm
from comment.utils import get_comment_from_key, get_user_for_request, CommentFailReason
from comment.mixins import CanCreateMixin, CanEditMixin, CanDeleteMixin, CommentCreateMixin, BaseCommentView
from comment.responses import UTF8JsonResponse
from comment.messages import EmailError


class CreateComment(CanCreateMixin, CommentCreateMixin):
    comment = None
    email_service = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['comment'] = self.comment
        return context

    def get_template_names(self):
        if self.request.user.is_anonymous or self.comment.is_parent:
            return 'comment/comments/base.html'
        else:
            return 'comment/comments/child_comment.html'

    def form_valid(self, form):
        user = get_user_for_request(self.request)
        comment_content = form.cleaned_data['content']
        email = form.cleaned_data.get('email', None)
        time_posted = timezone.now()
        temp_comment = Comment(
            content_object=self.model_obj,
            content=comment_content,
            user=user,
            parent=self.parent_comment,
            email=email or user.email,
            posted=time_posted
        )
        self.comment = self.perform_create(temp_comment, self.request)
        self.data = render_to_string(self.get_template_names(), self.get_context_data(), request=self.request)
        return UTF8JsonResponse(self.json())

    def form_invalid(self, form):
        self.error = EmailError.EMAIL_INVALID
        self.status = 400
        return UTF8JsonResponse(self.json(), status=self.status)


class UpdateComment(CanEditMixin, BaseCommentView):
    comment = None

    def get_object(self):
        self.comment = get_object_or_404(Comment, pk=self.kwargs.get('pk'))
        return self.comment

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        context['comment_form'] = CommentForm(instance=self.comment, request=self.request)
        context['comment'] = self.comment
        self.data = render_to_string('comment/comments/update_comment.html', context, request=self.request)
        return UTF8JsonResponse(self.json())

    def post(self, request, *args, **kwargs):
        form = CommentForm(request.POST, instance=self.comment, request=self.request)
        context = self.get_context_data()
        if form.is_valid():
            form.save()
            context['comment'] = self.comment
            self.data = render_to_string('comment/comments/comment_content.html', context, request=self.request)
            return UTF8JsonResponse(self.json())


class DeleteComment(CanDeleteMixin, BaseCommentView):
    comment = None

    def get_object(self):
        self.comment = get_object_or_404(Comment, pk=self.kwargs.get('pk'))
        return self.comment

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        context["comment"] = self.comment
        context['has_parent'] = not self.comment.is_parent
        self.data = render_to_string('comment/comments/comment_modal.html', context, request=request)
        return UTF8JsonResponse(self.json())

    def post(self, request, *args, **kwargs):
        self.comment.delete()
        context = self.get_context_data()
        self.data = render_to_string('comment/comments/base.html', context, request=self.request)
        return UTF8JsonResponse(self.json())


class ConfirmComment(CommentCreateMixin):

    @staticmethod
    def _handle_invalid_comment(comment, request):
        if comment.why_invalid == CommentFailReason.BAD:
            messages.error(request, EmailError.BROKEN_VERIFICATION_LINK)
        elif comment.why_invalid == CommentFailReason.EXISTS:
            messages.warning(request, EmailError.USED_VERIFICATION_LINK)

    def get(self, request, *args, **kwargs):
        key = kwargs.get('key', None)
        temp_comment = get_comment_from_key(key)
        self._handle_invalid_comment(temp_comment, request)

        if not temp_comment.is_valid:
            return render(request, template_name='comment/anonymous/discarded.html')

        comment = self.perform_save(temp_comment.obj, request)

        return redirect(comment.get_url(request))



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


def set_default_email(apps, schema_editor):
    comment_model = apps.get_model('comment', 'Comment')
    for comment in comment_model.objects.all():
        comment.email = comment.user.email
        comment.save(update_fields=['email'])


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('comment', '0008_comment_urlhash'),
    ]

    operations = [
        migrations.AddField(
            model_name='comment',
            name='email',
            field=models.EmailField(default=None, max_length=254, null=True),
        ),
        migrations.RunPython(set_default_email, migrations.RunPython.noop),
        migrations.AlterField(
            model_name='comment',
            name='email',
            field=models.EmailField(max_length=254, blank=True),
        ),
        migrations.AlterField(
            model_name='comment',
            name='posted',
            field=models.DateTimeField(default=django.utils.timezone.now, editable=False),
        ),
        migrations.AlterField(
            model_name='comment',
            name='user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.db import models


class FollowerManager(models.Manager):
    def is_following(self, email, model_object):
        content_type = ContentType.objects.get_for_model(model_object)
        return self.filter(email=email, object_id=model_object.id, content_type=content_type).exists()

    def follow(self, email, username, model_object):
        if not email or self.is_following(email, model_object):
            return None
        return self.create(email=email, username=username, content_object=model_object)

    def unfollow(self, email, model_object):
        content_type = ContentType.objects.get_for_model(model_object)
        self.get(email=email, object_id=model_object.id, content_type=content_type).delete()

    def toggle_follow(self, email, username, model_object):
        if not email:
            return False
        if self.is_following(email, model_object):
            self.unfollow(email, model_object)
            return False
        self.follow(email, username, model_object)
        return True

    def follow_parent_thread_for_comment(self, comment):
        """This method is used to set the comment's creator as a follower of own comment and the parent thread"""
        if not comment.email:
            return
        username = comment.get_username()
        model_object = comment
        if not comment.is_parent:
            model_object = comment.parent
        else:
            # follow the main thread for parent comment
            self.follow(comment.email, username, comment.content_object)
        self.follow(comment.email, username, model_object)

    def filter_for_model_object(self, model_obj):
        content_type = ContentType.objects.get_for_model(model_obj)
        return self.filter(content_type=content_type, object_id=model_obj.id)

    def get_emails_for_model_object(self, model_obj):
        return self.filter_for_model_object(model_obj).values_list('email', flat=True)



from django.utils.translation import gettext_lazy as _
from django.conf import settings

PROFILE_APP_NAME = None
PROFILE_MODEL_NAME = None
COMMENT_PROFILE_API_FIELDS = []
COMMENT_USER_API_FIELDS = ['id', 'username', 'email']

COMMENT_USE_GRAVATAR = False

COMMENT_FLAGS_ALLOWED = 0
COMMENT_SHOW_FLAGGED = False
COMMENT_FLAG_REASONS = [
    (1, _('Spam | Exists only to promote a service')),
    (2, _('Abusive | Intended at promoting hatred')),
]
COMMENT_URL_PREFIX = 'comment-'
COMMENT_URL_SUFFIX = ''
COMMENT_URL_ID_LENGTH = 8
COMMENT_PER_PAGE = 10

COMMENT_ALLOW_ANONYMOUS = False
COMMENT_FROM_EMAIL = getattr(settings, 'EMAIL_HOST_USER', None)   # used for sending confirmation emails
COMMENT_CONTACT_EMAIL = COMMENT_FROM_EMAIL  # used for contact address in confirmation emails
COMMENT_SEND_HTML_EMAIL = True
COMMENT_ANONYMOUS_USERNAME = 'Anonymous User'
COMMENT_USE_EMAIL_FIRST_PART_AS_USERNAME = False

COMMENT_ALLOW_TRANSLATION = False

COMMENT_ALLOW_SUBSCRIPTION = False

COMMENT_ORDER_BY = ['-posted']

COMMENT_WRAP_CONTENT_WORDS = 30

COMMENT_DEFAULT_PROFILE_PIC_LOC = '/static/img/default.png'



import re

from django import template

from comment.models import ReactionInstance, FlagInstance, Follower
from comment.forms import CommentForm
from comment.utils import (
    is_comment_moderator, is_comment_admin, get_gravatar_img, get_profile_instance, get_wrapped_words_number
)
from comment.managers import FlagInstanceManager
from comment.messages import ReactionError
from comment.context import DABContext

MULTIPLE_NEW_LINE_RE = re.compile(r'(.*)(\n){2,}(.*)')
SINGLE_NEW_LINE_RE = re.compile(r'(.*)(\n)(.*)')
register = template.Library()


@register.simple_tag(name='get_model_name')
def get_model_name(obj):
    """ returns the model name of an object """
    return type(obj).__name__


@register.simple_tag(name='get_app_name')
def get_app_name(obj):
    """ returns the app name of an object """
    return type(obj)._meta.app_label


@register.simple_tag(name='get_username_for_comment')
def get_username_for_comment(comment):
    return comment.get_username()


@register.simple_tag(name='get_profile_url')
def get_profile_url(obj):
    """ returns profile url of user """
    if not obj.user:
        return get_gravatar_img(obj.email)
    profile = get_profile_instance(obj.user)
    if profile:
        return profile.get_absolute_url()
    return get_gravatar_img(obj.email)


@register.simple_tag(name='get_img_path')
def get_img_path(obj):
    """ returns url of profile image of a user """
    profile = get_profile_instance(obj.user)
    if not profile:
        return get_gravatar_img(obj.email)
    for field in profile.__class__._meta.get_fields():
        if hasattr(field, 'upload_to'):
            return field.value_from_object(profile).url
    return get_gravatar_img(obj.email)


@register.simple_tag(name='get_comments_count')
def get_comments_count(obj, user):
    return obj.comments.all_comments_by_object(obj, include_flagged=is_comment_moderator(user)).count()


@register.simple_tag(name='get_comment_replies')
def get_comment_replies(comment, user):
    return comment.replies(include_flagged=is_comment_moderator(user))


@register.simple_tag(name='get_replies_count')
def get_replies_count(comment, user):
    return comment.replies(include_flagged=is_comment_moderator(user)).count()


def render_comments(obj, request, oauth=False):
    """
    Retrieves list of comment related to a certain object and renders the appropriate template
    """
    context = DABContext(request, model_object=obj)
    context.update({
        'comment_form': CommentForm(request=request),
        'oauth': oauth,
        'request': request,
    })
    return context


register.inclusion_tag('comment/base.html')(render_comments)


def render_content(comment, number=None):
    try:
        number = int(number)
    except (ValueError, TypeError):
        number = get_wrapped_words_number()

    # Restrict 2 or more line breaks to 2 <br>
    content = MULTIPLE_NEW_LINE_RE.sub(r'\1<br><br>\3', comment.content)
    content = SINGLE_NEW_LINE_RE.sub(r'\1<br>\3', content)
    content_words = content.split()
    if not number or len(content_words) <= number:
        text_1 = content
        text_2 = None
    else:
        text_1 = ' '.join(content_words[:number])
        text_2 = ' '.join(content_words[number:])

    return {
        'text_1': text_1,
        'text_2': text_2,
        'urlhash': comment.urlhash
    }


@register.simple_tag(name='can_delete_comment')
def can_delete_comment(comment, user):
    return is_comment_admin(user) or (comment.is_flagged and is_comment_moderator(user))


register.inclusion_tag('comment/comments/content.html')(render_content)


@register.simple_tag(name='include_static')
def include_static():
    """ This function shall be deprecated """
    return ''


def include_bootstrap():
    """ include static files """
    return None


register.inclusion_tag('comment/bootstrap.html')(include_bootstrap)


@register.simple_tag(name='render_field')
def render_field(field, **kwargs):
    field.field.widget.attrs.update(kwargs)
    return field


@register.simple_tag(name='has_reacted')
def has_reacted(comment, user, reaction):
    """
    Returns whether a user has reacted with a particular reaction on a comment or not.
    """
    if user.is_authenticated:
        reaction_type = getattr(ReactionInstance.ReactionType, reaction.upper(), None)
        if not reaction_type:
            raise template.TemplateSyntaxError(ReactionError.TYPE_INVALID.format(reaction_type=reaction))
        return ReactionInstance.objects.filter(
            user=user,
            reaction_type=reaction_type.value,
            reaction__comment=comment
            ).exists()

    return False


@register.filter(name='has_flagged')
def has_flagged(user, comment):
    if user.is_authenticated:
        return FlagInstance.objects.filter(user=user, flag__comment=comment).exists()

    return False


@register.filter(name='has_followed')
def has_followed(user, model_object):
    if user.is_authenticated:
        return Follower.objects.is_following(user.email, model_object)
    return False


@register.simple_tag(name='render_flag_reasons')
def render_flag_reasons():
    return FlagInstanceManager.reasons_list



from django.urls import path, re_path
from django.views.decorators.cache import cache_page
from django.views.i18n import JavaScriptCatalog

from comment import __version__
from comment.views import (
    CreateComment, UpdateComment, DeleteComment, SetReaction, SetFlag, ChangeFlagState,
    ConfirmComment, ToggleFollowView
)

app_name = 'comment'


urlpatterns = [
    path('create/', CreateComment.as_view(), name='create'),
    path('edit/<int:pk>/', UpdateComment.as_view(), name='edit'),
    path('delete/<int:pk>/', DeleteComment.as_view(), name='delete'),
    path('<int:pk>/react/<str:reaction>/', SetReaction.as_view(), name='react'),
    path('<int:pk>/flag/', SetFlag.as_view(), name='flag'),
    path('<int:pk>/flag/state/change/', ChangeFlagState.as_view(), name='flag-change-state'),
    re_path(r'^confirm/(?P<key>[^/]+)/$', ConfirmComment.as_view(), name='confirm-comment'),
    path('toggle-subscription/', ToggleFollowView.as_view(), name='toggle-subscription'),
    # javascript translations
    # The value returned by _get_version() must change when translations change.
    path(
        'jsi18n/',
        cache_page(86400, key_prefix='js18n-%s' % __version__)(JavaScriptCatalog.as_view()),
        name='javascript-catalog'),
]



from unittest.mock import patch

from rest_framework import status

from comment.conf import settings
from comment.messages import FlagInfo, FlagError
from comment.tests.base import BaseCommentFlagTest


class SetFlagViewTest(BaseCommentFlagTest):
    def setUp(self):
        super().setUp()
        self.flag_data.update({
            'info': ''
            })
        self.response_data = {
            'status': 1
        }

    def test_set_flag(self):
        _url = self.get_url('comment:flag', self.comment.id)
        self.flag_data['reason'] = 1
        response = self.client.post(_url, data=self.flag_data)

        response_data = {
            'data': {
                'status': 0,
                'flag': 1,
            },
            'anonymous': False,
            'error': None,
            'msg': FlagInfo.FLAGGED_SUCCESS
        }
        self.assertEqual(response.status_code, 200)
        server_response = response.json()
        self.assertDictEqual(server_response, response_data)
        self.assertTextTranslated(server_response['msg'], _url)

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 0)
    def test_set_flag_when_flagging_not_enabled(self):
        settings.COMMENT_FLAGS_ALLOWED = 0
        _url = self.get_url('comment:flag', self.comment.id)
        self.flag_data['reason'] = 1
        response = self.client.post(_url, data=self.flag_data)
        self.assertEqual(response.status_code, 403)

    def test_set_flag_for_flagging_old_comments(self):
        """Test backward compatibility for this update"""
        _url = self.get_url('comment:flag', self.comment.id)
        data = self.flag_data.copy()
        # delete the flag object
        self.comment.flag.delete()
        response = self.client.post(_url, data=data)
        response_data = {
            'data': {
                'status': 0,
                'flag': 1,
            },
            'anonymous': False,
            'error': None,
            'msg': FlagInfo.FLAGGED_SUCCESS
        }
        self.assertEqual(response.status_code, 200)
        server_response = response.json()
        self.assertDictEqual(server_response, response_data)
        self.assertTextTranslated(server_response['msg'], _url)

    def test_set_flag_for_unflagging(self):
        # un-flag => no reason is passed and the comment must be already flagged by the user
        _url = self.get_url('comment:flag', self.comment_2.id)
        data = {}
        response = self.client.post(_url, data=data)
        response_data = {
            'data': {
                'status': 0
            },
            'anonymous': False,
            'error': None,
            'msg': FlagInfo.UNFLAGGED_SUCCESS
        }
        self.assertEqual(response.status_code, 200)
        server_response = response.json()
        self.assertDictEqual(server_response, response_data)
        self.assertTextTranslated(server_response['msg'], _url)

    def test_set_flag_for_unauthenticated_user(self):
        """Test whether unauthenticated user can create/delete flag using view"""
        url = self.get_url('comment:flag', self.comment.id).replace('?', '')
        self.client.logout()
        response = self.client.post(url, data=self.flag_data)
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.url, '{}?next={}'.format(settings.LOGIN_URL, url))

    def test_get_request(self):
        """Test whether GET requests are allowed or not"""
        url = self.get_url('comment:flag', self.comment.id)
        response = self.client.get(url, data=self.flag_data)
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_non_ajax_requests(self):
        """Test response if non AJAX requests are sent"""
        url = self.get_url('comment:flag', self.comment.id)
        response = self.client_non_ajax.post(url, data=self.flag_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_incorrect_comment_id(self):
        """Test response when an incorrect comment id is passed"""
        url = self.get_url('comment:flag', 102_876)
        response = self.client.post(url, data=self.flag_data)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_incorrect_reason(self):
        """Test response when incorrect reason is passed"""
        url = self.get_url('comment:flag', self.comment.id)
        data = self.flag_data
        reason = -1
        data.update({'reason': reason})
        response = self.client.post(url, data=data)
        self.assertEqual(response.status_code, 400)


class ChangeFlagStateViewTest(BaseCommentFlagTest):
    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1)
    def setUp(self):
        super().setUp()
        self.data = {
            'state': self.comment_for_change_state.flag.REJECTED
        }
        self.create_flag_instance(self.user_1, self.comment_for_change_state, **self.flag_data)
        self.create_flag_instance(self.user_2, self.comment_for_change_state, **self.flag_data)

    def test_change_flag_state_for_unflagged_comment(self):
        comment = self.create_comment(self.content_object_1)
        self.assertFalse(comment.is_flagged)
        self.client.force_login(self.moderator)
        self.assertEqual(int(self.client.session['_auth_user_id']), self.moderator.id)
        response = self.client.post(self.get_url('comment:flag-change-state', comment.id), data=self.data)
        self.assertEqual(response.status_code, 400)

    def test_change_flag_state_by_not_permitted_user(self):
        self.assertTrue(self.comment_for_change_state.is_flagged)
        self.client.force_login(self.user_1)
        self.assertEqual(int(self.client.session['_auth_user_id']), self.user_1.id)
        response = self.client.post(
            self.get_url('comment:flag-change-state', self.comment_for_change_state.id), data=self.data
        )
        self.assertEqual(response.status_code, 403)

    def test_change_flag_state_with_wrong_state_value(self):
        self.assertTrue(self.comment_for_change_state.is_flagged)
        self.client.force_login(self.moderator)
        self.assertEqual(int(self.client.session['_auth_user_id']), self.moderator.id)
        self.assertEqual(self.comment_for_change_state.flag.state, self.comment_for_change_state.flag.FLAGGED)

        # valid state is REJECTED and RESOLVED
        self.data['state'] = self.comment_for_change_state.flag.UNFLAGGED
        response = self.client.post(
            self.get_url('comment:flag-change-state', self.comment_for_change_state.id), data=self.data
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json()['error'], FlagError.STATE_CHANGE_ERROR)
        self.assertEqual(self.comment_for_change_state.flag.state, self.comment_for_change_state.flag.FLAGGED)

    def test_change_flag_state_success(self):
        self.assertTrue(self.comment_for_change_state.is_flagged)
        self.client.force_login(self.moderator)
        self.assertEqual(int(self.client.session['_auth_user_id']), self.moderator.id)
        self.assertEqual(self.comment_for_change_state.flag.state, self.comment_for_change_state.flag.FLAGGED)

        # valid state is REJECTED and RESOLVED
        self.data['state'] = self.comment_for_change_state.flag.REJECTED
        response = self.client.post(
            self.get_url('comment:flag-change-state', self.comment_for_change_state.id), data=self.data
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()['data']['state'], self.comment_for_change_state.flag.REJECTED)
        self.comment_for_change_state.flag.refresh_from_db()
        self.assertEqual(self.comment_for_change_state.flag.moderator, self.moderator)
        self.assertEqual(self.comment_for_change_state.flag.state, self.comment_for_change_state.flag.REJECTED)



from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from django.db import models

from comment.managers.followers import FollowerManager


class Follower(models.Model):
    email = models.EmailField()
    username = models.CharField(max_length=50)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey()

    objects = FollowerManager()

    def __str__(self):
        return f'{str(self.content_object)} followed by {self.email}'

    def __repr__(self):
        return self.__str__()



from django.http import JsonResponse


class DABResponseData:
    status = 200
    data = None
    error = None
    anonymous = False
    msg = None

    def json(self):
        return {
            'data': self.data,
            'error': self.error,
            'anonymous': self.anonymous,
            'msg': self.msg
        }


class UTF8JsonResponse(JsonResponse):
    def __init__(self, *args, json_dumps_params=None, **kwargs):
        json_dumps_params = {"ensure_ascii": False, **(json_dumps_params or {})}
        super().__init__(*args, json_dumps_params=json_dumps_params, **kwargs)



from time import sleep
from unittest.mock import patch

from django.utils import timezone

from comment.conf import settings
from comment.models import Comment
from comment.tests.base import BaseCommentManagerTest


class CommentModelTest(BaseCommentManagerTest):
    def test_can_create_comment(self):
        parent_comment = self.create_comment(self.content_object_1)
        self.assertIsNotNone(parent_comment)
        self.assertEqual(str(parent_comment), f'comment by {parent_comment.user}: {parent_comment.content[:20]}')
        self.assertEqual(repr(parent_comment), f'comment by {parent_comment.user}: {parent_comment.content[:20]}')
        self.assertTrue(parent_comment.is_parent)
        self.assertEqual(parent_comment.replies().count(), 0)
        self.assertIsNotNone(parent_comment.urlhash)

        child_comment = self.create_comment(self.content_object_1, parent=parent_comment)
        self.assertIsNotNone(child_comment)
        self.assertEqual(str(child_comment), f'reply by {child_comment.user}: {child_comment.content[:20]}')
        self.assertEqual(repr(child_comment), f'reply by {child_comment.user}: {child_comment.content[:20]}')
        self.assertFalse(child_comment.is_parent)
        self.assertEqual(parent_comment.replies().count(), 1)
        self.assertIsNotNone(child_comment.urlhash)

    def test_is_edited(self):
        comment = self.create_comment(self.content_object_1)
        self.assertFalse(comment.is_edited)
        comment.content = 'updated'
        sleep(1)
        comment.save()
        self.assertTrue(comment.is_edited)

    def test_is_edited_for_anonymous_comment(self):
        comment = self.create_anonymous_comment(posted=timezone.now() - timezone.timedelta(days=1))
        self.assertFalse(comment.is_edited)

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1)
    @patch.object(settings, 'COMMENT_SHOW_FLAGGED', False)
    def test_replies_method(self):
        self.assertEqual(self.parent_comment_2.replies().count(), 3)
        reply = self.parent_comment_2.replies().first()
        self.create_flag_instance(self.user_1, reply)
        self.create_flag_instance(self.user_2, reply)
        # default replies method hides flagged comment
        self.assertEqual(self.parent_comment_2.replies().count(), 2)
        self.assertEqual(self.parent_comment_2.replies(include_flagged=True).count(), 3)

    @patch('comment.models.comments.hasattr')
    def test_is_flagged_property(self, mocked_hasattr):
        comment = self.create_comment(self.content_object_2)
        self.assertEqual(comment.flag.state, comment.flag.UNFLAGGED)
        self.assertFalse(comment.is_flagged)

        comment.flag.state = comment.flag.FLAGGED
        self.assertTrue(comment.is_flagged)

        with patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 0):
            self.assertIs(False, comment.is_flagged)

            mocked_hasattr.return_value = False
            self.assertIs(False, comment.is_flagged)

    @patch('comment.models.comments.hasattr')
    def test_has_flagged_state(self, mocked_hasattr):
        comment = self.create_comment(self.content_object_2)
        self.assertEqual(comment.flag.state, comment.flag.UNFLAGGED)
        self.assertFalse(comment.has_flagged_state)

        comment.flag.state = comment.flag.FLAGGED
        self.assertTrue(comment.has_flagged_state)

        mocked_hasattr.return_value = False
        self.assertFalse(comment.has_flagged_state)

    @patch('comment.models.comments.hasattr')
    def test_has_rejected_state(self, mocked_hasattr):
        comment = self.create_comment(self.content_object_2)
        self.assertEqual(comment.flag.state, comment.flag.UNFLAGGED)
        self.assertFalse(comment.has_rejected_state)

        comment.flag.state = comment.flag.REJECTED
        comment.flag.save()
        self.assertTrue(comment.has_rejected_state)

        mocked_hasattr.return_value = False
        self.assertFalse(comment.has_rejected_state)

    @patch('comment.models.comments.hasattr')
    def test_has_resolved_state(self, mocked_hasattr):
        comment = self.create_comment(self.content_object_2)
        self.assertEqual(comment.flag.state, comment.flag.UNFLAGGED)
        self.assertFalse(comment.has_resolved_state)

        comment.flag.state = comment.flag.RESOLVED
        comment.flag.save()
        self.assertTrue(comment.has_resolved_state)

        mocked_hasattr.return_value = False
        self.assertFalse(comment.has_resolved_state)

    @patch('comment.managers.CommentManager.generate_urlhash')
    def test_urlhash_is_unique(self, mocked_generate_urlhash):
        mocked_generate_urlhash.side_effect = ['first_urlhash', 'first_urlhash', 'second_urlhash']
        first_comment = self.create_comment(self.content_object_1)
        self.assertEqual(first_comment.urlhash, 'first_urlhash')
        mocked_generate_urlhash.assert_called_once()
        second_comment = self.create_comment(self.content_object_1)
        self.assertEqual(second_comment.urlhash, 'second_urlhash')
        self.assertEqual(mocked_generate_urlhash.call_count, 3)

    def test_comment_email(self):
        comment = self.parent_comment_1
        self.assertEqual(comment.email, comment.user.email)

    def test_get_url(self):
        from comment.tests.base import RequestFactory

        factory = RequestFactory()
        request = factory.get('/')
        request.user = self.user_1
        attr = 'COMMENT_PER_PAGE'
        comment = self.parent_comment_3

        # no pagination(parent comment 3, 2, 1 belong to the same content_object_1)
        with patch.object(settings, attr, 0):
            comment_url = comment.content_object.get_absolute_url() + '#' + comment.urlhash

            self.assertEqual(comment_url, comment.get_url(request))

        # with pagination
        with patch.object(settings, attr, 3):
            # comment on first page
            self.assertEqual(comment_url, comment.get_url(request))

            # comment on the second page
            comment = self.parent_comment_1
            comment_url = comment.content_object.get_absolute_url() + '?page=2' + '#' + comment.urlhash
            self.assertEqual(comment_url, comment.get_url(request))

    def test_get_username_for_non_anonymous_comment(self):
        comment = self.create_comment(self.content_object_1, user=self.user_1)

        self.assertEqual(comment.get_username(), comment.user.username)

    def test_get_username_for_anonymous_comment(self):
        comment = self.create_anonymous_comment()

        with patch.object(settings, 'COMMENT_USE_EMAIL_FIRST_PART_AS_USERNAME', False):
            self.assertEqual(comment.get_username(), settings.COMMENT_ANONYMOUS_USERNAME)

        with patch.object(settings, 'COMMENT_USE_EMAIL_FIRST_PART_AS_USERNAME', True):
            self.assertEqual(comment.get_username(), comment.email.split('@')[0])


class CommentModelManagerTest(BaseCommentManagerTest):

    class OrderForParentCommentsTest(BaseCommentManagerTest):
        def setUp(self):
            super().setUp()
            self.all_parents_qs = Comment.objects.all_exclude_flagged().filter(parent=None)
            self.all_comments_qs = Comment.objects.all_exclude_flagged()

        def test_default_value(self):
            self.assertQuerysetEqual(
                Comment.objects._filter_parents(self.all_comments_qs),
                self.all_parents_qs.order_by(*settings.COMMENT_ORDER_BY)
                )

        @patch.object(settings, 'COMMENT_ORDER_BY', ['-reaction__likes'])
        def test_custom_values(self):
            self.assertQuerysetEqual(
                Comment.objects._filter_parents(self.all_comments_qs),
                self.all_parents_qs.order_by(*settings.COMMENT_ORDER_BY)
                )

    def test_retrieve_all_parent_comments(self):
        # for all objects of a content type
        all_comments = Comment.objects.all().count()
        self.assertEqual(all_comments, 10)
        parent_comments = Comment.objects.all_parents().count()
        self.assertEqual(parent_comments, 5)

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1)
    def test_filtering_flagged_comment(self):
        comment = self.parent_comment_1
        self.assertEqual(Comment.objects.all_exclude_flagged().count(), self.increment)
        self.create_flag_instance(self.user_1, comment)
        self.create_flag_instance(self.user_2, comment)

        with patch.object(settings, 'COMMENT_SHOW_FLAGGED', False):
            self.assertEqual(Comment.objects.all_exclude_flagged().count(), self.increment - 1)

        with patch.object(settings, 'COMMENT_SHOW_FLAGGED', True):
            self.assertEqual(Comment.objects.all_exclude_flagged().count(), self.increment)

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 0)
    def test_filtering_comment_when_flag_not_enabled(self):
        comment = self.parent_comment_1
        self.assertEqual(Comment.objects.all_exclude_flagged().count(), self.increment)
        self.create_flag_instance(self.user_1, comment)
        self.create_flag_instance(self.user_2, comment)
        comment.flag.refresh_from_db()
        self.assertEqual(comment.flag.count, 2)
        self.assertEqual(Comment.objects.all_exclude_flagged().count(), self.increment)

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1)
    @patch.object(settings, 'COMMENT_SHOW_FLAGGED', False)
    def test_all_comments_by_object(self):
        # all comment for a particular content type
        init_count = self.post_1.comments.count()
        self.assertEqual(init_count, 6)

        comment = self.post_1.comments.first()
        self.create_flag_instance(self.user_1, comment)
        self.create_flag_instance(self.user_2, comment)
        # comments without flagged
        count = Comment.objects.all_comments_by_object(self.post_1).count()
        self.assertEqual(count, init_count - 1)

        # comments with flagged
        count = Comment.objects.all_comments_by_object(self.post_1, include_flagged=True).count()
        self.assertEqual(count, init_count)

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1)
    @patch.object(settings, 'COMMENT_SHOW_FLAGGED', False)
    def test_filter_parents_by_object(self):
        # parent comment only
        init_count = self.post_2.comments.filter(parent=None).count()
        self.assertEqual(init_count, 2)
        comment = Comment.objects.filter_parents_by_object(self.post_2).first()
        self.create_flag_instance(self.user_1, comment)
        self.create_flag_instance(self.user_2, comment)
        # comments without flagged
        count = Comment.objects.filter_parents_by_object(self.post_2).count()
        self.assertEqual(count, init_count - 1)

        # comments with flagged
        count = Comment.objects.filter_parents_by_object(self.post_2, include_flagged=True).count()
        self.assertEqual(count, init_count)

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1)
    @patch.object(settings, 'COMMENT_SHOW_FLAGGED', False)
    def test_get_parent_comment(self):
        # no parent_id passed from the url
        self.assertIsNone(Comment.objects.get_parent_comment(''))
        # no parent_id passed as 0
        self.assertIsNone(Comment.objects.get_parent_comment('0'))
        # no parent_id doesn't exist passed -> although this is highly unlikely as this will be handled by the mixin
        # but is useful for admin interface if required
        self.assertIsNone(Comment.objects.get_parent_comment(100))
        parent_comment = Comment.objects.get_parent_comment(1)
        self.assertIsNotNone(parent_comment)
        self.assertEqual(parent_comment, self.parent_comment_1)



from django.contrib.auth import get_user_model
from django.utils import timezone

from django.apps import apps
from django.db.models import ImageField

from rest_framework import serializers

from comment.conf import settings
from comment.models import Comment, Flag, Reaction
from comment.utils import get_user_for_request, get_profile_instance
from comment.messages import EmailError
from comment.mixins import CommentCreateMixin


def get_profile_model():
    profile_app_name = getattr(settings, 'PROFILE_APP_NAME', None)
    profile_model_name = getattr(settings, 'PROFILE_MODEL_NAME', None)
    if not profile_model_name or not profile_app_name:
        return None
    return apps.get_model(profile_app_name, profile_model_name)


def get_user_fields():
    user_model = get_user_model()
    fields = user_model._meta.get_fields()
    api_fields = list(settings.COMMENT_USER_API_FIELDS) + ['profile']
    api_fields = list(set(api_fields))
    for field in fields:
        if hasattr(field, "upload_to") and isinstance(field, ImageField):
            api_fields.append(field.name)
    return api_fields


class ProfileSerializerDAB(serializers.ModelSerializer):
    class Meta:
        model = get_profile_model()
        fields = getattr(settings, 'COMMENT_PROFILE_API_FIELDS', '__all__')


class UserSerializerDAB(serializers.ModelSerializer):
    profile = serializers.SerializerMethodField()

    class Meta:
        model = get_user_model()
        fields = get_user_fields()
        lookup_field = model.USERNAME_FIELD

    @staticmethod
    def get_profile(obj):
        profile = get_profile_instance(obj)
        if not profile:
            return None
        return ProfileSerializerDAB(profile).data


class BaseCommentSerializer(serializers.ModelSerializer):
    user = UserSerializerDAB(read_only=True)
    parent = serializers.SerializerMethodField()
    replies = serializers.SerializerMethodField()
    reply_count = serializers.SerializerMethodField()

    @staticmethod
    def get_parent(obj):
        if obj.parent:
            return obj.parent.id
        else:
            return None

    @staticmethod
    def get_replies(obj):
        if obj.is_parent:
            return CommentSerializer(obj.replies(), many=True).data
        else:
            return []

    @staticmethod
    def get_reply_count(obj):
        if obj.is_parent:
            return obj.replies().count()
        else:
            return 0

    @staticmethod
    def get_is_flagged(obj):
        return obj.is_flagged

    @staticmethod
    def get_flags(obj):
        return FlagSerializer(obj.flag).data

    @staticmethod
    def get_reactions(obj):
        return ReactionSerializer(obj.reaction).data


class CommentCreateSerializer(BaseCommentSerializer, CommentCreateMixin):
    class Meta:
        model = Comment
        fields = ('id', 'user', 'email', 'content', 'parent', 'posted', 'edited', 'reply_count', 'replies', 'urlhash')

    def __init__(self, *args, **kwargs):
        user = kwargs['context']['request'].user
        self.email_service = None
        if user.is_authenticated or not settings.COMMENT_ALLOW_ANONYMOUS:
            del self.fields['email']

        super().__init__(*args, **kwargs)

    @staticmethod
    def validate_email(value):
        if not value:
            raise serializers.ValidationError(EmailError.EMAIL_MISSING, code='required')
        return value.strip().lower()

    def create(self, validated_data):
        request = self.context['request']
        user = get_user_for_request(request)
        content = validated_data.get('content')
        email = validated_data.get('email')
        time_posted = timezone.now()

        temp_comment = Comment(
            content_object=self.context['model_obj'],
            content=content,
            user=user,
            parent=self.context['parent_comment'],
            email=email,
            posted=time_posted
        )
        return self.perform_create(temp_comment, request, api=True)


class CommentSerializer(BaseCommentSerializer):
    is_flagged = serializers.SerializerMethodField()
    flags = serializers.SerializerMethodField()
    reactions = serializers.SerializerMethodField()

    class Meta:
        model = Comment
        fields = (
            'id', 'user', 'email', 'content', 'parent', 'posted', 'edited', 'reply_count', 'replies', 'reactions',
            'is_flagged', 'flags', 'urlhash'
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        context = kwargs.get('context')
        reaction_update = False
        flag_update = False
        if context:
            reaction_update = context.get('reaction_update')
            flag_update = context.get('flag_update')
        if reaction_update or flag_update:
            self.fields['content'].read_only = True


class ReactionSerializer(serializers.ModelSerializer):
    users = serializers.SerializerMethodField()

    class Meta:
        model = Reaction
        fields = ('likes', 'dislikes', 'users')

    @staticmethod
    def get_users(obj):
        users = {'likes': [], 'dislikes': []}
        for instance in obj.reactions.all():
            user_info = {
                'id': instance.user.id,
                'username': instance.user.USERNAME_FIELD
            }

            if instance.reaction_type == instance.ReactionType.LIKE:
                users['likes'].append(user_info)
            else:
                users['dislikes'].append(user_info)
        return users


class FlagSerializer(serializers.ModelSerializer):
    reporters = serializers.SerializerMethodField()
    verbose_state = serializers.SerializerMethodField()

    class Meta:
        model = Flag
        fields = ('count', 'moderator', 'state', 'verbose_state', 'reporters')

    @staticmethod
    def get_reporters(obj):
        return [
            {
                'id': flag_instance.user.id,
                'username': flag_instance.user.USERNAME_FIELD
            }
            for flag_instance in obj.flags.all()
        ]

    @staticmethod
    def get_verbose_state(obj):
        return obj.get_verbose_state(obj.state)



from rest_framework import permissions

from comment.conf import settings
from comment.utils import is_comment_admin, is_comment_moderator


class IsOwnerOrReadOnly(permissions.BasePermission):
    """
    Custom permission to only allow owner of an object to edit it.
    """
    def has_object_permission(self, request, view, obj):
        # GET, HEAD or OPTIONS requests are SAFE_METHODS.
        if request.method in permissions.SAFE_METHODS:
            return True
        # PUT and DELETE permissions are allowed to the owner of the comment.
        if request.method == 'DELETE':  # comment admin can delete other users comments
            return any([
                is_comment_admin(request.user),
                obj.user == request.user,
                (obj.is_flagged and is_comment_moderator(request.user))
            ])
        return obj.user == request.user


class FlagEnabledPermission(permissions.BasePermission):
    """
    This will check if the COMMENT_FLAGS_ALLOWED is enabled
    """
    def has_permission(self, request, view):
        return bool(getattr(settings, 'COMMENT_FLAGS_ALLOWED', 0))


class CanChangeFlaggedCommentState(permissions.BasePermission):
    def has_permission(self, request, view):
        return is_comment_admin(request.user) or is_comment_moderator(request.user)

    def has_object_permission(self, request, view, obj):
        return obj.is_flagged and (is_comment_admin(request.user) or is_comment_moderator(request.user))


class SubscriptionEnabled(permissions.BasePermission):
    """
    This will check if the COMMENT_ALLOW_SUBSCRIPTION is enabled
    """
    def has_permission(self, request, view):
        return getattr(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)


class CanGetSubscribers(SubscriptionEnabled):
    def has_permission(self, request, view):
        if not super().has_permission(request, view):
            return False
        return is_comment_admin(request.user) or is_comment_moderator(request.user)



# Generated by Django 3.0.7 on 2020-06-20 12:59

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('comment', '0006_flag_flaginstance'),
    ]

    operations = [
        migrations.AddField(
            model_name='flag',
            name='moderator',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='flag',
            name='state',
            field=models.SmallIntegerField(choices=[(1, 'Unflagged'), (2, 'Flagged'), (3, 'Flag rejected by the moderator'), (4, 'Comment modified by the author')], default=1),
        ),
    ]



import abc

from django.contrib.auth.mixins import LoginRequiredMixin, AccessMixin
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseBadRequest, HttpResponseForbidden
from django.views.generic import FormView

from comment.conf import settings
from comment.utils import is_comment_admin, is_comment_moderator
from comment.validators import ValidatorMixin
from comment.messages import ErrorMessage, FlagError, FollowError
from comment.service.email import DABEmailService
from comment.messages import EmailInfo
from comment.forms import CommentForm
from comment.context import DABContext
from comment.responses import DABResponseData


class AJAXRequiredMixin:
    def dispatch(self, request, *args, **kwargs):
        if not request.META.get('HTTP_X_REQUESTED_WITH', None) == 'XMLHttpRequest':
            return HttpResponseBadRequest(ErrorMessage.NON_AJAX_REQUEST)
        return super().dispatch(request, *args, **kwargs)


class BasePermission(AJAXRequiredMixin):
    def has_permission(self, request):
        return True

    def has_object_permission(self, request, obj):
        return True


class BaseCommentMixin(LoginRequiredMixin, BasePermission):
    pass


class BaseCommentView(FormView, DABResponseData):
    form_class = CommentForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['comment_form'] = context.pop('form')
        # context.update(get_comment_context_data(self.request))
        context.update(DABContext(self.request))
        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.request
        return kwargs


class CommentCreateMixin(BaseCommentView):
    email_service = None

    def _initialize_email_service(self, comment, request):
        self.email_service = DABEmailService(comment, request)

    def _send_notification_to_followers(self, comment, request):
        if settings.COMMENT_ALLOW_SUBSCRIPTION:
            self._initialize_email_service(comment, request)
            self.email_service.send_notification_to_followers()

    def perform_save(self, comment, request):
        comment.save()
        self._send_notification_to_followers(comment, request)
        comment.refresh_from_db()
        return comment

    def _handle_anonymous(self, comment, request, api=False):
        self._initialize_email_service(comment, request)
        self.email_service.send_confirmation_request(api=api)
        self.anonymous = True
        self.msg = EmailInfo.CONFIRMATION_SENT

    def perform_create(self, comment, request, api=False):
        if settings.COMMENT_ALLOW_ANONYMOUS and not comment.user:
            self._handle_anonymous(comment, request, api)
        else:
            comment = self.perform_save(comment, request)
        return comment


class CanCreateMixin(BasePermission, AccessMixin, ValidatorMixin):
    def has_permission(self, request):
        return request.user.is_authenticated or settings.COMMENT_ALLOW_ANONYMOUS

    def dispatch(self, request, *args, **kwargs):
        if not self.has_permission(request):
            return self.handle_no_permission()
        return super().dispatch(request, *args, **kwargs)


class ObjectLevelMixin(BaseCommentMixin):
    @abc.abstractmethod
    def get_object(self):
        raise ImproperlyConfigured(
            ErrorMessage.METHOD_NOT_IMPLEMENTED.format(class_name=self.__class__.__name__, method_name='get_object()')
        )


class CanEditMixin(ObjectLevelMixin, ValidatorMixin, abc.ABC):
    def has_object_permission(self, request, obj):
        return request.user == obj.user

    def dispatch(self, request, *args, **kwargs):
        obj = self.get_object()
        if not self.has_object_permission(request, obj):
            return self.handle_no_permission()
        return super().dispatch(request, *args, **kwargs)


class CanDeleteMixin(ObjectLevelMixin, ValidatorMixin, abc.ABC):
    def has_object_permission(self, request, obj):
        return request.user == obj.user or is_comment_admin(request.user) \
                or (obj.is_flagged and is_comment_moderator(request.user))

    def dispatch(self, request, *args, **kwargs):
        obj = self.get_object()
        if not self.has_object_permission(request, obj):
            return self.handle_no_permission()
        return super().dispatch(request, *args, **kwargs)


class BaseFlagMixin(ObjectLevelMixin, abc.ABC):
    def dispatch(self, request, *args, **kwargs):
        if not getattr(settings, 'COMMENT_FLAGS_ALLOWED', False):
            return HttpResponseForbidden(FlagError.SYSTEM_NOT_ENABLED)
        return super().dispatch(request, *args, **kwargs)


class CanSetFlagMixin(BaseFlagMixin, abc.ABC):
    def has_object_permission(self, request, obj):
        """user cannot flag their own comment"""
        return obj.user != request.user

    def dispatch(self, request, *args, **kwargs):
        obj = self.get_object()
        if not self.has_object_permission(request, obj):
            self.handle_no_permission()
        return super().dispatch(request, *args, **kwargs)


class CanEditFlagStateMixin(BaseFlagMixin, abc.ABC):
    def has_permission(self, request):
        return is_comment_admin(request.user) or is_comment_moderator(request.user)

    def dispatch(self, request, *args, **kwargs):
        obj = self.get_object()
        if not obj.is_flagged:
            return HttpResponseBadRequest(FlagError.NOT_FLAGGED_OBJECT)
        if not self.has_permission(request):
            return self.handle_no_permission()
        return super().dispatch(request, *args, **kwargs)


class CanSubscribeMixin(BaseCommentMixin):
    def dispatch(self, request, *args, **kwargs):
        if not getattr(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False):
            return HttpResponseForbidden(FollowError.SYSTEM_NOT_ENABLED)
        return super().dispatch(request, *args, **kwargs)



import sys
from unittest.mock import patch

from rest_framework import status

from comment.exceptions import CommentBadRequest
from comment.tests.base import TestCase
from comment.messages import ExceptionError


class CommentExceptionTest(TestCase):
    _default_detail = ExceptionError.BAD_REQUEST

    def test_can_create_custom_error_without_params(self):
        exception = CommentBadRequest()

        self.assertEqual(exception.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(exception.detail, self._default_detail)

    def test_create_custom_error_with_params(self):
        detail = 'not found'
        exception = CommentBadRequest(detail=detail, status_code=404)

        self.assertEqual(exception.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(exception.detail, detail)

    def test_create_custom_error_without_drf_installed(self):
        with patch.dict(sys.modules, {'rest_framework.exceptions': None}):
            from importlib import reload
            reload(sys.modules['comment.exceptions'])
            from comment.exceptions import CommentBadRequest
            exception = CommentBadRequest()

        self.assertEqual(exception.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(exception.detail, self._default_detail)



from unittest.mock import patch

from django.core import mail
from django.test import RequestFactory

from comment.conf import settings
from comment.models import Comment, Follower
from comment.api.serializers import get_profile_model, get_user_fields, UserSerializerDAB, CommentCreateSerializer, \
    CommentSerializer
from comment.tests.test_api.test_views import APIBaseTest


class APICommentSerializers(APIBaseTest):
    def setUp(self):
        super().setUp()
        self.parent_count = Comment.objects.filter_parents_by_object(self.post_1).count()
        self.all_count = Comment.objects.all().count()

    def increase_count(self, parent=False):
        if parent:
            self.parent_count += 1
        self.all_count += 1

    def comment_count_test(self):
        self.assertEqual(Comment.objects.filter_parents_by_object(self.post_1).count(), self.parent_count)
        self.assertEqual(Comment.objects.all().count(), self.all_count)

    def test_get_profile_model(self):
        # missing settings attrs
        with patch.object(settings, 'PROFILE_APP_NAME', None):
            profile = get_profile_model()
            self.assertIsNone(profile)

        # providing wrong attribute value, an exception is raised
        with patch.object(settings, 'PROFILE_APP_NAME', 'wrong'):
            self.assertRaises(LookupError, get_profile_model)

        # attribute value is None
        with patch.object(settings, 'PROFILE_APP_NAME', None):
            profile = get_profile_model()
            self.assertIsNone(profile)

        # success
        with patch.object(settings, 'PROFILE_APP_NAME', 'user_profile'):
            profile = get_profile_model()
            self.assertIsNotNone(profile)

    def test_user_serializer(self):
        # PROFILE_MODEL_NAME not provided
        with patch.object(settings, 'PROFILE_MODEL_NAME', None):
            profile = UserSerializerDAB.get_profile(self.user_1)
            self.assertIsNone(profile)

        # PROFILE_MODEL_NAME is wrong
        with patch.object(settings, 'PROFILE_MODEL_NAME', 'wrong'):
            profile = UserSerializerDAB.get_profile(self.user_1)
            self.assertIsNone(profile)

        # success
        with patch.object(settings, 'PROFILE_MODEL_NAME', 'userprofile'):
            profile = UserSerializerDAB.get_profile(self.user_1)
            self.assertIsNotNone(profile)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', False)
    def test_create_parent_comment_serializer(self):
        self.assertEqual(self.parent_count, 3)
        self.assertEqual(self.all_count, 8)

        factory = RequestFactory()
        request = factory.get('/')
        request.user = self.user_1
        data = {
            'model_obj': self.post_1,
            'parent_comment': None,
            'request': request
        }

        serializer = CommentCreateSerializer(context=data)
        self.assertIsNone(serializer.fields.get('email'))
        comment = serializer.create(validated_data={'content': 'test'})
        self.increase_count(parent=True)
        self.comment_count_test()
        self.assertIsNotNone(comment)

        # get parent
        parent_id = serializer.get_parent(comment)
        self.assertIsNone(parent_id)

        # get replies
        replies = serializer.get_replies(comment)
        reply_count = serializer.get_reply_count(comment)
        self.assertEqual(replies, [])
        self.assertEqual(reply_count, 0)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', False)
    def test_create_child_comment_serializer(self):
        self.assertEqual(self.parent_count, 3)
        self.assertEqual(self.all_count, 8)

        factory = RequestFactory()
        request = factory.get('/')
        request.user = self.user_1
        data = {
            'model_obj': self.post_1,
            'request': request,
            'parent_comment': self.comment_1
        }

        serializer = CommentCreateSerializer(context=data)
        comment = serializer.create(validated_data={'content': 'test'})
        self.increase_count()
        self.comment_count_test()
        self.assertIsNotNone(comment)

        # get parent
        parent_id = CommentCreateSerializer.get_parent(comment)
        self.assertEqual(parent_id, data['parent_comment'].id)

        replies = serializer.get_replies(self.comment_1)
        reply_count = serializer.get_reply_count(self.comment_1)
        self.assertIsNotNone(replies)
        self.assertEqual(reply_count, 2)

        replies = serializer.get_replies(self.comment_4)
        reply_count = serializer.get_reply_count(self.comment_4)
        self.assertEqual(replies, [])
        self.assertEqual(reply_count, 0)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', False)
    def test_send_notification(self):
        factory = RequestFactory()
        request = factory.get('/')
        request.user = self.user_1
        data = {
            'model_obj': self.post_1,
            'request': request,
            'parent_comment': self.comment_1
        }

        Follower.objects.follow('e@e.com', 'testUser', self.comment_1)

        serializer = CommentCreateSerializer(context=data)
        comment = serializer.create(validated_data={'content': 'test'})
        self.assertTrue(serializer.email_service._email_thread.is_alive)
        self.assertIsNotNone(comment)
        self.assertIsNotNone(serializer.email_service._email_thread)
        serializer.email_service._email_thread.join()
        self.assertEqual(len(mail.outbox), 1)

    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', True)
    def test_create_comment_serializer_for_anonymous(self):
        from django.contrib.auth.models import AnonymousUser
        factory = RequestFactory()
        request = factory.get('/')
        request.user = AnonymousUser()
        data = {
            'model_obj': self.post_1,
            'parent_comment': None,
            'request': request
        }
        serializer = CommentCreateSerializer(context=data)

        self.assertIsNotNone(serializer.fields['email'])
        comment = serializer.create(validated_data={
            'content': 'anonymous posting',
            'email': 'abc@abc.com'
        })
        # no creation occurs until comment is verified
        self.comment_count_test()
        self.assertIsNotNone(comment)

        # confirmation email is sent
        self.assertIsNotNone(serializer.email_service._email_thread)
        serializer.email_service._email_thread.join()
        self.assertEqual(len(mail.outbox), 1)

    def test_passing_context_to_serializer(self):
        serializer = CommentSerializer(self.comment_1)
        self.assertFalse(serializer.fields['content'].read_only)

        serializer = CommentSerializer(self.comment_1, context={'reaction_update': True})
        self.assertTrue(serializer.fields['content'].read_only)

        serializer = CommentSerializer(self.comment_1, context={'flag_update': True})
        self.assertTrue(serializer.fields['content'].read_only)


class TestProfileSerializer(APIBaseTest):
    def test_default_fields(self):
        fields = get_user_fields()
        self.assertSetEqual(set(fields), set(settings.COMMENT_USER_API_FIELDS + ['profile']))

    @patch('comment.api.serializers.isinstance')
    @patch('comment.api.serializers.hasattr')
    def test_has_image_field(self, mocked_hasattr, mocked_isinstance):
        mocked_isinstance.return_value = True
        mocked_hasattr.return_value = True
        fields = get_user_fields()
        self.assertIs('logentry' in fields, True)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _
from django.db.models import signals


class CommentConfig(AppConfig):
    name = 'comment'
    verbose_name = _('comment')

    def ready(self):
        import comment.signals

        signals.post_migrate.connect(comment.signals.create_permission_groups, sender=self)
        signals.post_migrate.connect(comment.signals.adjust_flagged_comments, sender=self)



from django import forms

from comment.models import Comment
from comment.conf import settings
from comment.messages import EmailInfo


class CommentForm(forms.ModelForm):
    class Meta:
        model = Comment
        fields = ('content',)
        widgets = {'content': forms.Textarea(attrs={'rows': 1})}

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request')
        super().__init__(*args, **kwargs)
        if self.request.user.is_anonymous and settings.COMMENT_ALLOW_ANONYMOUS:
            self.fields['email'] = forms.EmailField(
                label=EmailInfo.LABEL,
                widget=forms.EmailInput(attrs={
                    'placeholder': EmailInfo.INPUT_PLACEHOLDER,
                    'title': EmailInfo.INPUT_TITLE
                    })
                )

    def clean_email(self):
        """this will only be executed when email field is present for unauthenticated users"""
        email = self.cleaned_data['email']
        return email.strip().lower()



from django.contrib.auth.models import Permission, Group
from django.contrib.contenttypes.models import ContentType

from comment.conf import settings
from comment.models import Comment


def create_permission_groups(sender, **kwargs):
    if settings.COMMENT_FLAGS_ALLOWED:
        comment_ct = ContentType.objects.get_for_model(Comment)
        delete_comment_perm, __ = Permission.objects.get_or_create(
            codename='delete_comment',
            name='Can delete comment',
            content_type=comment_ct
        )
        admin_group, __ = Group.objects.get_or_create(name='comment_admin')
        admin_group.permissions.add(delete_comment_perm)
        delete_flagged_comment_perm, __ = Permission.objects.get_or_create(
            codename='delete_flagged_comment',
            name='Can delete flagged comment',
            content_type=comment_ct
        )
        moderator_group, __ = Group.objects.get_or_create(name='comment_moderator')
        moderator_group.permissions.add(delete_flagged_comment_perm)
        admin_group.permissions.add(delete_flagged_comment_perm)


def adjust_flagged_comments(sender, **kwargs):
    if settings.COMMENT_FLAGS_ALLOWED:
        for comment in Comment.objects.all():
            comment.flag.toggle_flagged_state()



from django.urls import path, re_path
from rest_framework.urlpatterns import format_suffix_patterns

from comment.api import views

urlpatterns = [
    path('comments/', views.CommentList.as_view(), name='comments-list'),
    path('comments/create/', views.CommentCreate.as_view(), name='comments-create'),
    path('comments/<int:pk>/', views.CommentDetail.as_view(), name='comment-detail'),
    path('comments/<int:pk>/react/<str:reaction>/', views.CommentDetailForReaction.as_view(), name='comments-react'),
    path('comments/<int:pk>/flag/', views.CommentDetailForFlag.as_view(), name='comments-flag'),
    path(
        'comments/<int:pk>/flag/state/change/',
        views.CommentDetailForFlagStateChange.as_view(),
        name='comments-flag-state-change'
    ),
    re_path(r'^comments/confirm/(?P<key>[^/]+)/$', views.ConfirmComment.as_view(), name='confirm-comment'),
    path('comments/toggle-subscription/', views.ToggleFollowAPI.as_view(), name='toggle-subscription'),
    path('comments/subscribers/', views.SubscribersAPI.as_view(), name='subscribers'),
]

urlpatterns = format_suffix_patterns(urlpatterns)



from unittest.mock import patch

from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import Page

from comment.conf import settings

from comment.tests.base import BaseCommentUtilsTest
from comment.context import DABContext
from comment.messages import ErrorMessage


class DABContextTest(BaseCommentUtilsTest):
    def setUp(self):
        super().setUp()
        self.data = {
            'model_object': self.post_1,
            'model_name': 'post',
            'model_id': self.post_1.id,
            'app_name': 'post',
            'user': self.post_1.author,
            'page': 10,
        }
        self.request = self.factory.post('/', data=self.data)
        self.request.user = self.post_1.author

    def test_initialized_object_is_dict(self):
        self.assertIsInstance(DABContext(self.request), dict)

    def test_model_object_exist_even_if_not_provided(self):
        self.assertEqual(DABContext(self.request).model_object, self.post_1)

    @patch.object(settings, 'LOGIN_URL', None)
    def test_get_login_url_fails_when_missing_in_settings(self):
        with self.assertRaises(ImproperlyConfigured) as e:
            DABContext(self.request)
        self.assertEqual(e.exception.args[0], ErrorMessage.LOGIN_URL_MISSING)

    @patch.object(settings, 'LOGIN_URL', 'profile/login/')
    def test_get_login_url_prepend_slash(self):
        self.assertFalse(settings.LOGIN_URL.startswith('/'))
        self.assertTrue(DABContext(self.request).get('login_url').startswith('/'))

    @patch.object(settings, 'LOGIN_URL', '/profile/login/')
    def test_get_login_url_success(self):
        self.assertTrue(DABContext(self.request).get('login_url'), '/profile/login/')

    def test_is_oauth_true(self):
        self.data['oauth'] = True
        request = self.factory.post('/', data=self.data)
        request.user = self.post_1.author
        context = DABContext(request)
        self.assertTrue(context['oauth'])

    def test_is_oauth_false(self):
        self.data['oauth'] = False
        request = self.factory.post('/', data=self.data)
        request.user = self.post_1.author
        context = DABContext(request)
        self.assertFalse(context['oauth'])

    @patch.object(settings, 'COMMENT_PER_PAGE', 0)
    def test_get_comments_without_pagination(self):
        context = DABContext(self.request)
        self.assertEqual(context['comments'].count(), self.increment)
        self.assertNotIsInstance(context['comments'], Page)

    @patch.object(settings, 'COMMENT_PER_PAGE', 2)
    def test_get_comments_with_pagination(self):
        context = DABContext(self.request)
        self.assertEqual(context['comments'].paginator.per_page, 2)
        self.assertIsInstance(context['comments'], Page)

    def test_context_object_is_callable(self):
        context = DABContext(self.request)
        self.assertTrue(callable(context))

    def test_calling_text_object_return_default_values(self):
        context = DABContext(self.request, extra_1='test-1')
        self.assertEqual(len(context), len(context.__call__()) + 1)
        self.assertEqual(len(context()), len(context.__call__()))



# Generated by Django 3.1.1 on 2020-10-23 14:42

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('comment', '0009_auto_20200811_1945'),
    ]

    operations = [
        migrations.AlterField(
            model_name='flag',
            name='moderator',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='flags_moderated', to=settings.AUTH_USER_MODEL),
        ),
    ]



from django.urls import reverse

from comment.tests.base import BaseCommentMixinTest
from comment.views import BaseToggleFollowView
from comment.messages import FollowError, EmailError
from comment.models import Follower


class BaseToggleFollowViewTest(BaseCommentMixinTest):
    def setUp(self):
        super().setUp()
        self.view = BaseToggleFollowView()
        self.toggle_follow_url = self.get_url(
            reverse('toggle-subscription'), app_name='comment', model_name='comment', model_id=self.comment.id
        )

    def test_assertion_error_on_missing_request_class(self):
        self.assertRaises(AssertionError, self.view.get_response_class)

    def test_success_on_providing_request_class(self):
        self.view.response_class = 'test'
        self.assertEqual(self.view.get_response_class(), self.view.response_class)

    def test_email_required_to_follow_object(self):
        self.client.force_login(self.user_without_email)
        response = self.client.post(self.toggle_follow_url, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json()['error']['email_required'], FollowError.EMAIL_REQUIRED.format(model_object=self.comment)
        )

    def test_invalid_email(self):
        data = {'email': 'invalid_email'}
        response = self.client.post(self.toggle_follow_url, HTTP_X_REQUESTED_WITH='XMLHttpRequest', data=data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json()['error']['invalid_email'], EmailError.EMAIL_INVALID)

    def test_toggle_follow_for_valid_email(self):
        self.client.force_login(self.user_without_email)
        self.assertEqual(self.user_without_email.email, '')
        data = {'email': 't@t.com'}
        self.assertFalse(Follower.objects.is_following(data['email'], self.comment))
        response = self.client.post(self.toggle_follow_url, HTTP_X_REQUESTED_WITH='XMLHttpRequest', data=data)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.json()['data']['following'])
        self.assertEqual(response.json()['data']['model_id'], self.comment.id)
        self.assertEqual(response.wsgi_request.user.email, data['email'])
        self.assertEqual(response.wsgi_request.user, self.user_without_email)

    def test_provide_email_for_user_has_already_email(self):
        self.client.force_login(self.user_2)
        self.assertEqual(self.user_2.email, 'test-2@acme.edu')
        data = {'email': 't@t.com'}
        self.assertFalse(Follower.objects.is_following(data['email'], self.comment))
        response = self.client.post(self.toggle_follow_url, HTTP_X_REQUESTED_WITH='XMLHttpRequest', data=data)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.json()['data']['following'])
        self.assertEqual(response.json()['data']['model_id'], self.comment.id)
        self.assertNotEqual(self.user_2, data['email'])
        self.assertEqual(response.wsgi_request.user, self.user_2)



from time import sleep
from unittest.mock import patch

from django.test import RequestFactory
from django.core import signing, mail

from rest_framework import status

from comment.conf import settings
from comment.models import Comment, FlagInstanceManager
from comment.messages import ContentTypeError, EmailError
from comment.api.serializers import CommentSerializer
from comment.api.permissions import (
    IsOwnerOrReadOnly, FlagEnabledPermission, CanChangeFlaggedCommentState, SubscriptionEnabled,
    CanGetSubscribers)
from comment.api.views import CommentList
from comment.tests.base import BaseCommentTest, timezone
from comment.tests.test_utils import BaseAnonymousCommentTest
from comment.utils import is_comment_admin, is_comment_moderator


class APIBaseTest(BaseCommentTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.comment_1 = cls.create_comment(cls.content_object_1)
        cls.comment_2 = cls.create_comment(cls.content_object_1)
        cls.comment_3 = cls.create_comment(cls.content_object_1)
        cls.comment_4 = cls.create_comment(cls.content_object_1, parent=cls.comment_1)
        cls.reaction_1 = cls.create_reaction_instance(cls.user_1, cls.comment_1, 'like')

        cls.comment_5 = cls.create_comment(cls.content_object_2)
        cls.comment_6 = cls.create_comment(cls.content_object_2)
        cls.comment_7 = cls.create_comment(cls.content_object_2, parent=cls.comment_5)
        cls.comment_8 = cls.create_comment(cls.content_object_2, parent=cls.comment_5)
        cls.reaction_2 = cls.create_reaction_instance(cls.user_1, cls.comment_5, 'dislike')

    def setUp(self):
        super().setUp()
        self.addCleanup(patch.stopall)


class APIPermissionsTest(APIBaseTest):
    def setUp(self):
        super().setUp()
        self.owner_permission = IsOwnerOrReadOnly()
        self.flag_enabled_permission = FlagEnabledPermission()
        self.can_change_flagged_comment_state = CanChangeFlaggedCommentState()
        self.subscription_enabled = SubscriptionEnabled()
        self.get_subscribers_permission = CanGetSubscribers()
        self.factory = RequestFactory()
        self.view = CommentList()

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.flag_data = {
            'reason': FlagInstanceManager.reason_values[0],
            'info': '',
        }
        cls.create_flag_instance(cls.user_1, cls.comment_1, **cls.flag_data)
        cls.create_flag_instance(cls.user_2, cls.comment_1, **cls.flag_data)

    def test_owner_permission(self):
        # self.client.login(username='test-2', password='1234')
        request = self.factory.get('/')
        # get is in the safe methods
        self.assertTrue(self.owner_permission.has_object_permission(request, self.view, self.comment_1))

        # PUT method from different user
        request = self.factory.put('/')
        request.user = self.user_2
        self.assertEqual(self.comment_1.user, self.user_1)
        self.assertFalse(self.owner_permission.has_object_permission(request, self.view, self.comment_1))

        # DELETE method from admin
        request = self.factory.put('/')
        request.user = self.admin
        self.assertEqual(self.comment_1.user, self.user_1)
        self.assertFalse(self.owner_permission.has_object_permission(request, self.view, self.comment_1))

        # PUT method from same user
        request = self.factory.put('/')
        request.user = self.user_1
        self.assertEqual(self.comment_1.user, self.user_1)
        self.assertTrue(self.owner_permission.has_object_permission(request, self.view, self.comment_1))

    def test_flag_enabled_permission(self):
        request = self.factory.get('/')
        with patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 0):
            self.assertIs(False, self.flag_enabled_permission.has_permission(request, self.view))
        with patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1):
            self.assertIs(True, self.flag_enabled_permission.has_permission(request, self.view))

    def test_can_change_flagged_comment_state(self):
        request = self.factory.get('/')
        user = self.user_1
        request.user = user  # not moderator user
        self.assertFalse(self.can_change_flagged_comment_state.has_permission(request, self.view))

        user = self.moderator
        request.user = user
        self.assertTrue(self.can_change_flagged_comment_state.has_permission(request, self.view))

        comment = self.comment_2
        self.assertFalse(comment.is_flagged)
        self.assertFalse(
            self.can_change_flagged_comment_state.has_object_permission(request, self.view, comment)
        )
        with patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1):
            self.set_flag(self.user_1, comment, **self.flag_data)
            self.set_flag(self.user_2, comment, **self.flag_data)
            self.assertIs(True, comment.is_flagged)
            self.assertIs(
                True,
                self.can_change_flagged_comment_state.has_object_permission(request, self.view, comment)
            )

            request.user = self.user_1
            self.assertIs(
                False,
                self.can_change_flagged_comment_state.has_object_permission(request, self.view, comment)
            )

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_normal_users_cannot_retrieve_subscribers(self):
        request = self.factory.get('/')
        request.user = self.user_1
        self.assertFalse(is_comment_admin(self.user_1))
        self.assertFalse(is_comment_moderator(self.user_1))
        self.assertTrue(settings.COMMENT_ALLOW_SUBSCRIPTION)
        self.assertFalse(self.get_subscribers_permission.has_permission(request, self.view))

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_only_moderators_can_retrieve_subscribers(self):
        request = self.factory.get('/')
        request.user = self.moderator
        self.assertTrue(is_comment_moderator(self.moderator))
        self.assertTrue(settings.COMMENT_ALLOW_SUBSCRIPTION)
        self.assertTrue(self.get_subscribers_permission.has_permission(request, self.view))

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    def test_cannot_retrieve_subscribers_when_system_disabled(self):
        request = self.factory.get('/')
        request.user = self.moderator
        self.assertTrue(is_comment_moderator(self.moderator))
        self.assertFalse(settings.COMMENT_ALLOW_SUBSCRIPTION)
        self.assertFalse(self.get_subscribers_permission.has_permission(request, self.view))

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    def test_cannot_subscribe_when_system_disabled(self):
        request = self.factory.post('/')
        self.assertFalse(settings.COMMENT_ALLOW_SUBSCRIPTION)
        self.assertFalse(self.subscription_enabled.has_permission(request, self.view))

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_can_subscribe(self):
        request = self.factory.post('/')
        self.assertTrue(settings.COMMENT_ALLOW_SUBSCRIPTION)
        self.assertTrue(self.subscription_enabled.has_permission(request, self.view))


class APICommentViewsTest(APIBaseTest):
    def setUp(self):
        super().setUp()
        self.url_data = {
            'model_name': 'post',
            'app_name': 'post',
            'model_id': 1
        }
        self.comment_count = Comment.objects.filter_parents_by_object(self.post_1).count()
        self.all_comments = Comment.objects.all().count()

    @staticmethod
    def get_url(base_url=None, **kwargs):
        if not base_url:
            base_url = '/api/comments/'
        if kwargs:
            base_url += '?'
            for (key, val) in kwargs.items():
                base_url += str(key) + '=' + str(val) + '&'
        return base_url.rstrip('&')

    def increase_count(self, parent=False):
        if parent:
            self.comment_count += 1
        self.all_comments += 1

    def comment_count_test(self):
        self.assertEqual(Comment.objects.filter_parents_by_object(self.post_1).count(), self.comment_count)
        self.assertEqual(Comment.objects.all().count(), self.all_comments)

    def test_can_retrieve_comments_list(self):
        response = self.client.get(self.get_url(**self.url_data))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 3)  # 3 parent comment, child comment will be nested in the parent.

    def test_retrieving_comment_list_without_app_name(self):
        data = self.url_data.copy()
        data.pop('app_name')
        url = self.get_url(**data)

        response = self.client.get(url)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['detail'], ContentTypeError.APP_NAME_MISSING)
        self.assertTextTranslated(response.data['detail'], url)

    def test_retrieving_comment_list_without_model_name(self):
        data = self.url_data.copy()
        data.pop('model_name')
        url = self.get_url(**data)

        response = self.client.get(url)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['detail'], ContentTypeError.MODEL_NAME_MISSING)
        self.assertTextTranslated(response.data['detail'], url)

    def test_retrieving_comment_list_without_model_id(self):
        url_data = self.url_data.copy()
        url_data.pop('model_id')
        url = self.get_url(**url_data)

        response = self.client.get(url)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['detail'], ContentTypeError.MODEL_ID_MISSING)
        self.assertTextTranslated(response.data['detail'], url)

    def test_retrieving_comment_list_with_invalid_app_name(self):
        data = self.url_data.copy()
        app_name = 'invalid'
        data['app_name'] = app_name
        url = self.get_url(**data)

        response = self.client.get(url)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['detail'], ContentTypeError.APP_NAME_INVALID.format(app_name=app_name))
        self.assertTextTranslated(response.data['detail'], url)

    def test_retrieving_comment_list_with_invalid_model_name(self):
        # not exist model type
        url_data = self.url_data.copy()
        model_name = 'does_not_exists'
        url_data['model_name'] = model_name
        url = self.get_url(**url_data)
        response = self.client.get(url)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['detail'], ContentTypeError.MODEL_NAME_INVALID.format(model_name=model_name))
        self.assertTextTranslated(response.data['detail'], url)

    def test_retrieving_comment_list_with_invalid_model_id(self):
        # not exist model id
        url_data = self.url_data.copy()
        model_id = 100
        url_data['model_id'] = model_id
        url = self.get_url(**url_data)
        response = self.client.get(url)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.data['detail'],
            ContentTypeError.MODEL_ID_INVALID.format(model_id=model_id, model_name=url_data["model_name"])
        )
        self.assertTextTranslated(response.data['detail'], url)

        # not integer model id
        model_id = 'c'
        url_data['model_id'] = model_id
        url = self.get_url(**url_data)
        response = self.client.get(url)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.data['detail'],
            ContentTypeError.ID_NOT_INTEGER.format(var_name='model', id=model_id)
        )
        self.assertTextTranslated(response.data['detail'], url)

    def test_create_parent_comment(self):
        # create parent comment
        self.assertEqual(self.comment_count, 3)
        self.assertEqual(self.all_comments, 8)

        base_url = '/api/comments/create/'
        data = {'content': 'new parent comment from api'}
        url_data = self.url_data.copy()

        response = self.client.post(self.get_url(base_url, **url_data), data=data)
        self.assertEqual(response.status_code, 201)
        comment_id = response.json()['id']
        # test email in database for authenticated user
        self.assertEqual(Comment.objects.get(id=comment_id).email, response.wsgi_request.user.email)
        self.increase_count(parent=True)
        self.comment_count_test()

    def test_create_child_comment(self):
        # create parent comment
        self.assertEqual(self.comment_count, 3)
        self.assertEqual(self.all_comments, 8)

        base_url = '/api/comments/create/'
        url_data = self.url_data.copy()
        # create child comment
        url_data['parent_id'] = 1
        data = {'content': 'new child comment from api'}

        response = self.client.post(self.get_url(base_url, **url_data), data=data)
        self.assertEqual(response.status_code, 201)
        self.increase_count()
        self.comment_count_test()

        # create comment with parent value = 0
        url_data['parent_id'] = 0
        data = {'content': 'new comment from api'}
        response = self.client.post(self.get_url(base_url, **url_data), data=data)
        self.assertEqual(response.status_code, 201)
        self.increase_count(parent=True)
        self.comment_count_test()

    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', True)
    def test_create_comment_for_anonymous_with_invalid_data(self):
        base_url = '/api/comments/create/'
        url_data = self.url_data.copy()

        # test anonymous commenting
        self.client.logout()

        # test invalid data
        data = {'content': 'new anonymous comment from api', 'email': ''}
        url = self.get_url(base_url, **url_data)
        response = self.client.post(url, data=data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json()['email'], [EmailError.EMAIL_MISSING])
        self.assertTextTranslated(response.json()['email'][0], base_url)

    @patch.object(settings, 'COMMENT_ALLOW_ANONYMOUS', True)
    def test_create_comment_for_anonymous_with_valid_data(self):
        base_url = '/api/comments/create/'
        url_data = self.url_data.copy()
        url = self.get_url(base_url, **url_data)

        # test anonymous commenting
        self.client.logout()

        # test valid data
        data = {'content': 'new anonymous comment from api', 'email': 'a@a.com'}

        response = self.client.post(url, data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        # no count should change
        self.comment_count_test()

    def test_cannot_create_child_comment(self):
        # parent id not integer
        base_url = '/api/comments/create/'
        url_data = self.url_data.copy()
        parent_id = 'c'
        url_data['parent_id'] = parent_id
        data = {'content': 'new child comment from api'}
        response = self.client.post(self.get_url(base_url, **url_data), data=data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.data['detail'],
            ContentTypeError.ID_NOT_INTEGER.format(var_name='parent', id=parent_id)
        )
        self.assertTextTranslated(response.data['detail'], base_url)

        # parent id not exist
        parent_id = 100
        url_data['parent_id'] = parent_id
        data = {'content': 'new child comment from api'}
        response = self.client.post(self.get_url(base_url, **url_data), data=data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['detail'], ContentTypeError.PARENT_ID_INVALID.format(parent_id=parent_id))
        self.assertTextTranslated(response.data['detail'], base_url)

        # parent id doesn't belong to the model object
        parent_id = 1
        url_data.update({
            'parent_id': parent_id,
            'model_id': 2
        })
        data = {'content': 'new child comment from api'}
        response = self.client.post(self.get_url(base_url, **url_data), data=data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['detail'], ContentTypeError.PARENT_ID_INVALID.format(parent_id=parent_id))

    def test_can_retrieve_update_delete_comment(self):
        count = Comment.objects.all().count()
        self.assertEqual(count, 8)
        # retrieve
        response = self.client.get('/api/comments/2/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['id'], 2)
        self.assertEqual(response.data['content'], 'comment 2')

        # update
        data = {'content': 'updated comment'}
        response = self.client.put('/api/comments/2/', data=data, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['id'], 2)
        self.assertEqual(response.data['content'], data['content'])
        self.assertEqual(count, Comment.objects.all().count())

        # delete
        response = self.client.delete('/api/comments/2/')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Comment.objects.all().count(), count - 1)

        # delete parent will delete its children as well
        count = Comment.objects.all().count()
        self.assertEqual(count, 7)
        comment = Comment.objects.get(id=1)
        self.assertEqual(comment.replies().count(), 1)
        response = self.client.delete('/api/comments/1/')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Comment.objects.all().count(), count - 2)

    def test_react_to_comment_success(self):
        # post like - comment has no reaction
        self.assertEqual(self.comment_3.likes, 0)
        self.assertEqual(self.comment_3.dislikes, 0)
        response = self.client.post(f'/api/comments/{self.comment_3.id}/react/like/')
        self.assertEqual(response.status_code, 200)
        self.comment_3.reaction.refresh_from_db()
        self.assertEqual(self.comment_3.likes, 1)
        self.assertEqual(self.comment_3.dislikes, 0)

        # post dislike - comment is liked by the user
        response = self.client.post(f'/api/comments/{self.comment_3.id}/react/dislike/')
        self.assertEqual(response.status_code, 200)
        self.comment_3.reaction.refresh_from_db()
        self.assertEqual(self.comment_3.likes, 0)
        self.assertEqual(self.comment_3.dislikes, 1)

        # post dislike - comment is disliked by the user => comment reaction is removed
        response = self.client.post(f'/api/comments/{self.comment_3.id}/react/dislike/')
        self.assertEqual(response.status_code, 200)
        self.comment_3.reaction.refresh_from_db()
        self.assertEqual(self.comment_3.likes, 0)
        self.assertEqual(self.comment_3.dislikes, 0)

    def test_react_to_comment_with_invalid_reaction_type(self):
        response = self.client.post(f'/api/comments/{self.comment_3.id}/react/invalid_type/')
        self.assertEqual(response.status_code, 400)


class APICommentFlagViewTest(APIBaseTest):

    def get_url(self, c_id=None):
        if not c_id:
            c_id = self.comment.id

        return f'/api/comments/{c_id}/flag/'

    def setUp(self):
        super().setUp()
        self.comment = self.comment_1
        self.user = self.user_1
        self.flag_data = {
            'reason': FlagInstanceManager.reason_values[0],
            'info': '',
        }

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.flag = cls.create_flag_instance(cls.user_2, cls.comment_2)

    def test_flag_to_comment(self):
        comment = self.comment
        data = self.flag_data
        # flag - comment has no flags
        url = self.get_url()
        self.assertEqual(comment.flag.count, 0)
        response = self.client.post(url, data=data)
        self.assertEqual(response.status_code, 200)
        self.assertFalse(response.json()['is_flagged'])
        comment.flag.refresh_from_db()
        self.assertEqual(comment.flag.count, 1)

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 0)
    def test_flag_comment_when_flagging_not_enabled(self):
        comment = self.comment
        data = self.flag_data
        url = self.get_url()
        self.assertEqual(comment.flag.count, 0)
        response = self.client.post(url, data=data)
        self.assertEqual(response.status_code, 403)

    def test_unflag_to_comment(self):
        comment = self.comment_2
        user = self.user_2
        self.client.force_login(user)
        url = self.get_url(comment.id)
        # un-flag - comment is flagged by the user and no reason is passed
        response = self.client.post(url)
        self.assertEqual(response.status_code, 200)
        comment.flag.refresh_from_db()
        self.assertEqual(comment.flag.count, 0)

    def test_flag_to_previous_comments(self):
        """Maintains backward compatibility"""
        comment = self.comment
        url = self.get_url()
        data = self.flag_data
        comment.flag.delete()   # delete the flag object
        response = self.client.post(url, data=data)
        self.assertEqual(response.status_code, 200)
        comment.refresh_from_db()
        self.assertEqual(comment.flag.count, 1)

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1)
    def test_is_flagged_value(self):
        # test is_flagged property
        comment = self.create_comment(self.content_object_1)
        data = self.flag_data
        url = self.get_url(comment.id)
        response = self.client.post(url, data=data)

        self.assertEqual(response.status_code, 200)
        self.assertIs(False, response.json()['is_flagged'])

        comment.flag.refresh_from_db()

        self.assertEqual(comment.flag.count, 1)
        self.client.force_login(self.user_2)

        response = self.client.post(url, data=data)

        self.assertEqual(response.status_code, 200)
        self.assertIs(True, response.json()['is_flagged'])

        comment.flag.refresh_from_db()

        self.assertEqual(comment.flag.count, 2)

    def test_flag_to_comment_with_invalid_reason(self):
        data = self.flag_data
        data.update({'reason': -1})
        response = self.client.post(self.get_url(), data=data)
        self.assertEqual(response.status_code, 400)


class APICommentDetailForFlagStateChangeTest(APIBaseTest):
    def setUp(self):
        super().setUp()
        self.client.force_login(self.moderator)
        self.data = {
            'state': self.comment_1.flag.REJECTED
        }

    @classmethod
    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 1)
    def setUpTestData(cls):
        super().setUpTestData()
        cls.flag_data = {
            'reason': FlagInstanceManager.reason_values[0],
            'info': None,
        }
        # flag comment_1
        cls.create_flag_instance(cls.user_1, cls.comment_1, **cls.flag_data)
        cls.create_flag_instance(cls.user_2, cls.comment_1, **cls.flag_data)

    def get_url(self, c_id=None):
        if not c_id:
            c_id = self.comment_1.id
        return f'/api/comments/{c_id}/flag/state/change/'

    @patch.object(settings, 'COMMENT_FLAGS_ALLOWED', 0)
    def test_change_state_when_flagging_is_disabled(self):
        response = self.client.post(self.get_url(), data=self.data)
        self.assertEqual(response.status_code, 403)

    def test_change_state_when_comment_is_not_flagged(self):
        comment = self.comment_2
        self.assertIs(False, comment.is_flagged)
        response = self.client.post(self.get_url(comment.id), data=self.data)
        self.assertEqual(response.status_code, 400)

    def test_change_state_by_not_permitted_user(self):
        comment = self.comment_1
        self.assertIs(True, comment.is_flagged)
        self.client.force_login(self.user_1)
        response = self.client.post(self.get_url(), data=self.data)
        self.assertEqual(response.status_code, 403)

    def test_change_state_with_wrong_state_value(self):
        data = self.data.copy()
        comment = self.comment_1
        self.assertIs(True, comment.is_flagged)

        data['state'] = 100
        response = self.client.post(self.get_url(), data=data)
        self.assertEqual(response.status_code, 400)

        data['state'] = "Not Int"
        response = self.client.post(self.get_url(), data=data)
        self.assertEqual(response.status_code, 400)

        response = self.client.post(self.get_url(), data={})
        self.assertEqual(response.status_code, 400)

        data['state'] = comment.flag.RESOLVED
        comment.refresh_from_db()
        self.assertFalse(comment.is_edited)
        response = self.client.post(self.get_url(), data=data)
        self.assertEqual(response.status_code, 400)

    def test_change_state_success(self):
        comment = self.comment_1
        self.assertIs(True, comment.is_flagged)
        self.assertEqual(comment.flag.state, comment.flag.FLAGGED)

        data = self.data.copy()
        data['state'] = comment.flag.REJECTED
        response = self.client.post(self.get_url(), data=data)
        self.assertEqual(response.status_code, 200)
        comment.flag.refresh_from_db()
        self.assertEqual(comment.flag.state, comment.flag.REJECTED)

        sleep(1)
        comment.content = "new content"
        comment.save()
        self.assertIs(True, comment.is_edited)

        # First request
        data['state'] = comment.flag.RESOLVED
        response = self.client.post(self.get_url(), data=data)
        self.assertEqual(response.status_code, 200)
        comment.flag.refresh_from_db()
        self.assertEqual(comment.flag.state, comment.flag.RESOLVED)

        # Second request of same state changes state to FLAGGED
        data['state'] = comment.flag.RESOLVED
        response = self.client.post(self.get_url(), data=data)
        self.assertEqual(response.status_code, 200)
        comment.flag.refresh_from_db()
        self.assertEqual(comment.flag.state, comment.flag.FLAGGED)


class APIConfirmCommentViewTest(BaseAnonymousCommentTest, APIBaseTest):
    def setUp(self):
        super().setUp()
        self.client.logout()
        self.init_count = Comment.objects.all().count()

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.comment = cls.create_anonymous_comment(posted=timezone.now(), email='a@a.com')

    def get_url(self, key=None):
        if not key:
            key = self.key

        return f'/api/comments/confirm/{key}/'

    def test_bad_signature(self):
        key = self.key + 'invalid'
        url = self.get_url(key)
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json()['detail'], EmailError.BROKEN_VERIFICATION_LINK)
        self.assertTextTranslated(response.data['detail'], url)
        self.assertEqual(Comment.objects.all().count(), self.init_count)

    def test_comment_exists(self):
        comment_dict = self.comment_obj.to_dict().copy()
        init_count = self.init_count
        comment_dict.update({
            'posted': str(self.comment.posted),
            'email': self.comment.email
        })
        key = signing.dumps(comment_dict)
        url = self.get_url(key)

        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['detail'], EmailError.USED_VERIFICATION_LINK)
        self.assertTextTranslated(response.data['detail'], url)
        self.assertEqual(Comment.objects.all().count(), init_count)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    def test_success_without_notification(self):
        response = self.client.get(self.get_url())
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        comment = Comment.objects.get(email=self.comment_obj.email, posted=self.time_posted)
        self.assertEqual(response.data, CommentSerializer(comment).data)
        self.assertEqual(Comment.objects.all().count(), self.init_count + 1)
        self.assertEqual(len(mail.outbox), 0)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_success_with_notification(self):
        response = self.client.get(self.get_url())
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        response.renderer_context['view'].email_service._email_thread.join()
        self.assertEqual(len(mail.outbox), 1)


class APIToggleFollowTest(APIBaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.comment_toggle_follow = cls.create_comment(cls.content_object_1)
        cls.app_name = cls.comment_toggle_follow._meta.app_label
        cls.model_name = cls.comment_toggle_follow.__class__.__name__
        cls.model_id = cls.comment_toggle_follow.id

    def get_url(self):
        params = [f'app_name={self.app_name}', f'model_name={self.model_name}', f'model_id={self.model_id}']
        return f'/api/comments/toggle-subscription/?{"&".join(params)}'

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_unauthenticated_users(self):
        self.client.logout()
        response = self.client.post(self.get_url())
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    def test_system_is_not_enabled(self):
        response = self.client.post(self.get_url())
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_toggle_follow(self):
        self.client.force_login(self.user_2)
        self.assertIsNotNone(self.user_2.email)
        response = self.client.post(self.get_url())
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json().get('data')
        self.assertEqual(data['app_name'], self.app_name)
        self.assertEqual(data['model_name'], self.model_name)
        self.assertEqual(data['model_id'], self.model_id)
        self.assertTrue(data['following'])


class APIGetSubscribersTest(APIBaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.comment_get_followers = cls.create_comment(cls.content_object_1)
        cls.app_name = cls.comment_get_followers._meta.app_label
        cls.model_name = cls.comment_get_followers.__class__.__name__
        cls.model_id = cls.comment_get_followers.id

    def get_url(self):
        params = [f'app_name={self.app_name}', f'model_name={self.model_name}', f'model_id={self.model_id}']
        return f'/api/comments/subscribers/?{"&".join(params)}'

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_unauthenticated_users(self):
        self.client.logout()
        response = self.client.get(self.get_url())
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', False)
    def test_system_is_not_enabled(self):
        response = self.client.get(self.get_url())
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_only_moderators_can_get_followers(self):
        self.client.force_login(self.moderator)
        response = self.client.get(self.get_url())
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['app_name'], self.app_name)
        self.assertEqual(response.data['model_name'], self.model_name)
        self.assertEqual(response.data['model_id'], self.model_id)
        # creator of comment, follow their comment automatically
        self.assertEqual(list(response.data['followers']), [self.comment_get_followers.email])

    @patch.object(settings, 'COMMENT_ALLOW_SUBSCRIPTION', True)
    def test_normal_users_cannot_get_followers(self):
        self.client.force_login(self.user_1)
        self.assertFalse(is_comment_moderator(self.user_1))
        response = self.client.get(self.get_url())
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)



from threading import Thread

from django.contrib.sites.shortcuts import get_current_site
from django.core import signing
from django.core.mail import EmailMultiAlternatives, get_connection
from django.template import loader
from django.urls import reverse

from comment.conf import settings
from comment.messages import EmailInfo
from comment.models import Follower


class DABEmailService(object):
    def __init__(self, comment, request):
        self.comment = comment
        self.request = request
        self.sender = settings.COMMENT_FROM_EMAIL
        self.is_html = settings.COMMENT_SEND_HTML_EMAIL
        self._email_thread = None

    def get_msg_context(self, **context):
        context['comment'] = self.comment
        context['site'] = get_current_site(self.request)
        context['contact'] = settings.COMMENT_CONTACT_EMAIL
        return context

    def get_message(self, subject, body, receivers, html_msg=None):
        msg = EmailMultiAlternatives(subject, body, self.sender, receivers)
        if html_msg:
            msg.attach_alternative(html_msg, 'text/html')
        return msg

    def send_messages(self, messages):
        connection = get_connection()  # Use default email connection
        self._email_thread = Thread(target=connection.send_messages, args=(messages,))
        self._email_thread.start()

    def get_message_templates(self, text_template, html_template, msg_context):
        text_msg_template = loader.get_template(text_template)
        text_msg = text_msg_template.render(msg_context)
        html_msg = None
        if self.is_html:
            html_msg_template = loader.get_template(html_template)
            html_msg = html_msg_template.render(msg_context)
        return text_msg, html_msg

    def send_confirmation_request(self, api=False):
        comment_dict = self.comment.to_dict()
        receivers = [comment_dict['email']]
        key = signing.dumps(comment_dict, compress=True)
        text_template = 'comment/anonymous/confirmation_request.txt'
        html_template = 'comment/anonymous/confirmation_request.html'
        subject = EmailInfo.CONFIRMATION_SUBJECT
        if api:
            confirmation_url = f'/api/comments/confirm/{key}/'
        else:
            confirmation_url = reverse('comment:confirm-comment', args=[key])

        context = self.get_msg_context(confirmation_url=confirmation_url)
        text_msg, html_msg = self.get_message_templates(text_template, html_template, context)
        msg = self.get_message(subject, text_msg, receivers, html_msg=html_msg)
        self.send_messages([msg])

    def get_thread(self):
        if self.comment.is_parent:
            return self.comment.content_object
        return self.comment.parent

    def get_thread_name(self):
        if self.comment.is_parent:
            return str(self.comment.content_object)
        return str(self.comment.parent).split(':')[0]

    def get_subject_for_notification(self, thread_name):
        username = self.comment.get_username()
        return EmailInfo.NOTIFICATION_SUBJECT.format(username=username, thread_name=thread_name)

    def get_messages_for_notification(self, thread_name, receivers):
        text_template = 'comment/notifications/notification.txt'
        html_template = 'comment/notifications/notification.html'
        subject = self.get_subject_for_notification(thread_name)
        messages = []
        for receiver in receivers:
            context = self.get_msg_context(thread_name=thread_name, receiver=receiver.username)
            text_msg, html_msg = self.get_message_templates(text_template, html_template, context)
            messages.append(self.get_message(subject, text_msg, [receiver.email], html_msg=html_msg))
        return messages

    def send_notification_to_followers(self):
        thread = self.get_thread()
        followers = Follower.objects.filter_for_model_object(thread).exclude(email=self.comment.email)
        if not followers:
            return
        thread_name = self.get_thread_name()
        messages = self.get_messages_for_notification(thread_name, followers)
        self.send_messages(messages)
